import pytest
from pathlib import Path, PurePath
from unittest.mock import patch
import aikido_zen.sinks.os

kind = "path_traversal"


def test_ospath_commands():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import os

        os.path.realpath("test/test2")
        op = "os.path.realpath"
        args = ("test/test2",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.path.getsize("aqkqjefbkqlleq_qkvfjksaicuaviel")
        op = "os.path.getsize"
        args = ("aqkqjefbkqlleq_qkvfjksaicuaviel",)
        # Need to use assert_any_call, since python 3.12 it uses os.path.join
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        os.path.realpath(b"te2st/test2")
        op = "os.path.realpath"
        args = (b"te2st/test2",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        path1 = Path("./", "../", "test/../test2")
        with pytest.raises(FileNotFoundError):
            os.path.getsize(path1)

        op = "os.path.getsize"
        args = (path1,)
        # Need to use assert_any_call, since python 3.12 it uses os.path.join
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)


def test_ospath_command_absolute_path():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import os

        os.path.abspath("../test/test2")
        op = "os.path.join"
        args = ("../test/test2",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        path1 = Path("./test", "test2", "test3")
        os.path.abspath(path1)

        op = "os.path.join"
        args = ("test/test2/test3",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)


def test_ospath_expanduser():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import os

        os.path.expanduser("../test/test2")
        op = "os.path.expanduser"
        args = ("../test/test2",)
        # Need to use assert_any_call, since python 3.12 it uses os.path.join
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        path1 = Path("./test", "test2", "test3")
        os.path.expanduser(path1)

        op = "os.path.expanduser"
        args = (path1,)
        # Need to use assert_any_call, since python 3.12 it uses os.path.join
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)


def test_ospath_expandvars():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import os

        os.path.expandvars("../test/test2")
        op = "os.path.expandvars"
        args = ("../test/test2",)
        # Need to use assert_any_call, since python 3.12 it uses os.path.join
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        path1 = Path("./test", "test2", "test3")
        os.path.expandvars(path1)

        op = "os.path.expandvars"
        args = (path1,)
        # Need to use assert_any_call, since python 3.12 it uses os.path.join
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)


def test_ospath_join():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import os

        os.path.join("../", "/etc/passwd", "..")
        op = "os.path.join"
        args1 = ("../",)
        args2 = ("/etc/passwd",)
        args3 = ("..",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args1)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args2)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args3)


@patch("aikido_zen.vulnerabilities.run_vulnerability_scan")
def test_ospath_join_bytes(mock_run_vulnerability_scan):
    import os

    op = "os.path.join"

    # Test with bytes arguments
    os.path.join(b"../", b"/etc/passwd", b"..")
    args1 = (b"../",)
    args2 = (b"/etc/passwd",)
    args3 = (b"..",)

    mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args1)
    mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args2)
    mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args3)


@patch("aikido_zen.vulnerabilities.run_vulnerability_scan")
def test_ospath_join_path_objects(mock_run_vulnerability_scan):
    import os

    op = "os.path.join"

    # Test with Path objects
    os.path.join(Path("../"), Path("/etc/passwd"), Path(".."))
    args1 = (Path("../"),)
    args2 = (Path("/etc/passwd"),)
    args3 = (Path(".."),)

    mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args1)
    mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args2)
    mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args3)


@patch("aikido_zen.vulnerabilities.run_vulnerability_scan")
def test_ospath_join_mixed_paths(mock_run_vulnerability_scan):
    import os

    op = "os.path.join"

    # Test with mixed strings and Path objects
    os.path.join("../", Path("/etc/passwd"), "..")
    args1 = ("../",)
    args2 = (Path("/etc/passwd"),)
    args3 = ("..",)

    mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args1)
    mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args2)
    mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args3)


def test_os_commands():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import os

        os.access(".xyzxyz", 0)
        op = "os.access"
        args = (".xyzxyz",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.chmod("1234567", 0)
        op = "os.chmod"
        args = ("1234567",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.chown("fLKEFLKENGKWBGKJEBLKALKKnkjfkj_jefkjwbgkjrw", 0, 0)
        op = "os.chown"
        args = ("fLKEFLKENGKWBGKJEBLKALKKnkjfkj_jefkjwbgkjrw",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        os.mkdir("qlkgkjbnlzheioe_kjbfkjeiLJ", 0)
        op = "os.mkdir"
        args = ("qlkgkjbnlzheioe_kjbfkjeiLJ",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.listdir("TEST_PATH_test")
        op = "os.listdir"
        args = ("TEST_PATH_test",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.readlink("Pathy_jgyr138")
        op = "os.readlink"
        args = ("Pathy_jgyr138",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.unlink("test_path.pathy")
        op = "os.unlink"
        args = ("test_path.pathy",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.unlink(b"wjlewjrlke")
        op = "os.unlink"
        args = (b"wjlewjrlke",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        os.rename("qlkgkjbnlzheioe_kjbfkjeiLJ", "lkflkenlnlgksnk_aknflkenfk")
        op = "os.rename"
        args = ("qlkgkjbnlzheioe_kjbfkjeiLJ",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)
        args = ("lkflkenlnlgksnk_aknflkenfk",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.rename(b"akflaflkqkajlqjoiq", b"kfjlfehfkakj")
        op = "os.rename"
        args = (b"akflaflkqkajlqjoiq",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)
        args = (b"kfjlfehfkakj",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.rename("akflaflkqkajlqjoiq", b"aflkkelwwgw")
        op = "os.rename"
        args = ("akflaflkqkajlqjoiq",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)
        args = (b"aflkkelwwgw",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        os.rmdir("lkflkenlnlgksnk_aknflkenfk")
        op = "os.rmdir"
        args = ("lkflkenlnlgksnk_aknflkenfk",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.remove("qlkgkjbnlzheioe_kjbfkjeiLJ")
        op = "os.remove"
        args = ("qlkgkjbnlzheioe_kjbfkjeiLJ",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        os.symlink("fqppnfqdbsklclfkn_lqbfjqbkwnd", "aqwfkjqkjqwfhoie_kjfhejlwqk")
        op = "os.symlink"
        args = ("fqppnfqdbsklclfkn_lqbfjqbkwnd",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)
        args = ("aqwfkjqkjqwfhoie_kjfhejlwqk",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)
        os.remove("aqwfkjqkjqwfhoie_kjfhejlwqk")

        path1 = PurePath("kkjehrqlknl_qk3rqlrjgna")
        with pytest.raises(FileNotFoundError):
            os.link("qkfhwifqqlejfke_qlfjboqvdbshjw", path1)
        op = "os.link"
        args = ("qkfhwifqqlejfke_qlfjboqvdbshjw",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)
        args = (path1,)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        os.walk("qlfjwqqfqelfknef_kwlkgrlkbwkalwd")
        op = "os.walk"
        args = ("qlfjwqqfqelfknef_kwlkgrlkbwkalwd",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            os.open("gmlwgewlnqlfnawglkwnfleknlew", os.O_RDONLY)
        op = "os.open"
        args = ("gmlwgewlnqlfnawglkwnfleknlew",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)


# Test that it can Handle all sorts of data :
def test_os_invalid_input():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import os

        op = "os.link"

        with pytest.raises(TypeError):
            os.link()
        mock_run_vulnerability_scan.assert_not_called()
        with pytest.raises(TypeError):
            os.link(123456789123456789)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called
        with pytest.raises(TypeError):
            os.link(None)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called
        with pytest.raises(TypeError):
            os.link(["list", "of", "commands"])
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test dictionary command
        with pytest.raises(TypeError):
            os.link({"key": "value"})
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test float command
        with pytest.raises(TypeError):
            os.link(3.14)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test boolean command
        with pytest.raises(TypeError):
            os.link(True)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test tuple command
        with pytest.raises(TypeError):
            os.link(("tuple", "command"))
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called


# Test that it can Handle all sorts of data :
def test_ospath_invalid_input():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import os

        op = "os.path.realpath"

        with pytest.raises(TypeError):
            os.path.realpath()
        mock_run_vulnerability_scan.assert_not_called()
        with pytest.raises(TypeError):
            os.path.realpath(123456789123456789)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called
        with pytest.raises(TypeError):
            os.path.realpath(None)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called
        with pytest.raises(TypeError):
            os.path.realpath(["list", "of", "commands"])
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test dictionary command
        with pytest.raises(TypeError):
            os.path.realpath({"key": "value"})
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test float command
        with pytest.raises(TypeError):
            os.path.realpath(3.14)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test boolean command
        with pytest.raises(TypeError):
            os.path.realpath(True)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test tuple command
        with pytest.raises(TypeError):
            os.path.realpath(("tuple", "command"))
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called



"""
Exports different types of Aikido Exception
"""

from aikido_zen.helpers.attack_human_name import attack_human_name


def generate_default_message(kind):
    """Generates a default message based on kind of attack"""
    return "Zen has blocked " + attack_human_name(kind)


class AikidoException(Exception):
    """General AikidoException class"""

    kind = None

    def __init__(self, message=None):
        if isinstance(message, str):
            super().__init__(self, message)
        else:
            super().__init__(self, generate_default_message(self.kind))


class AikidoSQLInjection(AikidoException):
    """Exception because of SQL Injection"""

    def __init__(self, dialect="unknown"):
        super().__init__(
            generate_default_message("sql_injection") + ", dialect: " + dialect
        )


class AikidoNoSQLInjection(AikidoException):
    """Exception because of NoSQL Injection"""

    kind = "nosql_injection"


class AikidoRateLimiting(AikidoException):
    """Exception caused when a page was rate limited"""

    def __init__(self, message="You are rate limited by Zen."):
        super().__init__(message)
        self.message = message


class AikidoShellInjection(AikidoException):
    """Exception becausen of Shell Injection"""

    kind = "shell_injection"


class AikidoPathTraversal(AikidoException):
    """Exception because of a path traversal"""

    kind = "path_traversal"


class AikidoSSRF(AikidoException):
    """Exception because of SSRF"""

    kind = "ssrf"



import pytest
from .is_redirect_status_code import is_redirect_status_code


def test_is_redirect_status_code():
    assert is_redirect_status_code(302) == True
    assert is_redirect_status_code(307) == True

    assert is_redirect_status_code(200) == False
    assert is_redirect_status_code(404) == False
    assert is_redirect_status_code(500) == False
    assert is_redirect_status_code(400) == False
    assert is_redirect_status_code(100) == False



"""Helper function file, see function docstring"""


def attack_human_name(kind):
    """Return the human name of each kind of attack"""
    if kind == "nosql_injection":
        return "a NoSQL injection"
    if kind == "sql_injection":
        return "an SQL injection"
    if kind == "shell_injection":
        return "a shell injection"
    if kind == "path_traversal":
        return "a path traversal attack"
    if kind == "ssrf":
        return "a server-side request forgery"
    return "unknown"



"""
Helper function file, see function doc for explanation
"""

import string

LOWERCASE = list(string.ascii_lowercase)
UPPERCASE = list(map(str.upper, LOWERCASE))
NUMBERS = list(string.digits)
SPECIAL = list("!#$%^&*|;:<>")
KNOWN_WORD_SEPARATORS = ["-"]
WHITE_SPACE = " "
MINIMUM_LENGTH = 10


def looks_like_a_secret(s):
    """
    This will make a judgement based on the string s wether
    or not s looks like a secret
    """
    if len(s) <= MINIMUM_LENGTH:
        return False

    has_number = any(char in s for char in NUMBERS)
    if not has_number:
        return False

    has_lower = any(char in s for char in LOWERCASE)
    has_upper = any(char in s for char in UPPERCASE)
    has_special = any(char in s for char in SPECIAL)
    charsets = [has_lower, has_upper, has_special]

    if charsets.count(True) < 2:
        return False

    if WHITE_SPACE in s:
        return False

    if any(separator in s for separator in KNOWN_WORD_SEPARATORS):
        return False

    window_size = MINIMUM_LENGTH
    ratios = []
    for i in range(len(s) - window_size + 1):
        window = s[i : i + window_size]
        unique_chars = set(window)
        ratios.append(len(unique_chars) / window_size)

    average_ratio = sum(ratios) / len(ratios)

    return average_ratio > 0.75



"""Exports `check_context_for_path_traversal`"""

from aikido_zen.helpers.extract_strings_from_context import extract_strings_from_context
from aikido_zen.helpers.try_parse_url import try_parse_url
from aikido_zen.helpers.path_to_string import path_to_string
from .detect_path_traversal import detect_path_traversal


def check_context_for_path_traversal(
    filename, operation, context, check_path_start=True
):
    """
    This will check the context for path traversal
    """
    is_url = try_parse_url(filename) is not None
    path_string = path_to_string(filename)
    if not path_string:
        return {}

    for user_input, path, source in extract_strings_from_context(context):
        if detect_path_traversal(path_string, user_input, check_path_start, is_url):
            return {
                "operation": operation,
                "kind": "path_traversal",
                "source": source,
                "pathToPayload": path,
                "metadata": {"filename": path_string},
                "payload": user_input,
            }
    return {}



"""
Helper function file, see funtion definition
"""

import re


def escape_string_regexp(string):
    """
    Escape characters with special meaning either inside or outside character sets.
    Use a simple backslash escape when it's always valid, and a '\\xnn' escape
    when the simpler form would be disallowed by Unicode patterns' stricter grammar.
    Taken from https://github.com/sindresorhus/escape-string-regexp/
    """
    pattern = re.compile(r"[|\\{}()[\]^$+*?.]")
    replace1 = re.sub(pattern, r"\\\g<0>", string)
    return re.sub(r"-", r"\\x2d", replace1)



"""
Exports ReportingApiHTTPRatelimited
"""

import aikido_zen.background_process.api.http_api as http_api
import aikido_zen.helpers.get_current_unixtime_ms as t


class ReportingApiHTTPRatelimited(http_api.ReportingApiHTTP):
    """HTTP Reporting API that has ratelimiting support"""

    def __init__(self, reporting_url, max_events_per_interval, interval_in_ms):
        super().__init__(reporting_url)
        self.interval_in_ms = interval_in_ms
        self.max_events_per_interval = max_events_per_interval
        self.events = []

    def report(self, token, event, timeout_in_sec):
        if event["type"] == "detected_attack":
            # Remove all outdated events :
            current_time = t.get_unixtime_ms()

            def event_in_interval_filter(e):
                return e["time"] > current_time - self.interval_in_ms

            self.events = list(filter(event_in_interval_filter, self.events))

            # Check if interval is exceeded :
            if len(self.events) >= self.max_events_per_interval:
                return {"success": False, "error": "max_attacks_reached"}

            self.events.append(event)
        return super().report(token, event, timeout_in_sec)



"""Exports create_interval"""


def create_interval(event_scheduler, interval_in_secs, function, args):
    """
    This function creates an interval which first runs "function"
    after waiting "interval_in_secs" seconds and after that keeps
    executing the function every "interval_in_secs" seconds.
    """
    # Sleep interval_in_secs seconds before starting the loop :
    event_scheduler.enter(
        interval_in_secs,
        1,
        interval_loop,
        (event_scheduler, interval_in_secs, function, args),
    )


def interval_loop(event_scheduler, interval_in_secs, function, args):
    """
    This is the actual interval loop which executes and schedules the function
    """
    # Execute function :
    function(*args)
    # Schedule the execution of the function in interval_in_secs seconds :
    event_scheduler.enter(
        interval_in_secs,
        1,
        interval_loop,
        (event_scheduler, interval_in_secs, function, args),
    )



"""
Mostly exports the class RateLimiter
"""

from aikido_zen.helpers.get_current_unixtime_ms import get_unixtime_ms
from .lru_cache import LRUCache


class RateLimiter:
    """
    Stores x times a request has been
    """

    def __init__(self, max_items, time_to_live_in_ms):
        self.max_items = max_items
        self.time_to_live_in_ms = time_to_live_in_ms
        self.rate_limited_items = LRUCache(max_items, time_to_live_in_ms)

    def is_allowed(self, key, window_size_in_ms, max_requests):
        """
        Checks if the request is allowed given the history
        """
        current_time = get_unixtime_ms()
        request_timestamps = self.rate_limited_items.get(key) or list()

        # clear entries that are not in the rate-limiting window anymore
        request_timestamps = [
            timestamp
            for timestamp in request_timestamps
            if timestamp >= current_time - window_size_in_ms
        ]

        # ensure the number of entries exceeds max_requests by only 1
        while len(request_timestamps) > (max_requests + 1):
            # We remove the oldest entry, since this one has become useless if the limit is already exceeded
            request_timestamps.pop(0)  # Remove the first element

        # update entries by adding the new timestamp and storing it in the LRU Cache
        request_timestamps.append(current_time)
        self.rate_limited_items.set(key, request_timestamps)

        # if the total amount of requests in the current window exceeds max requests, we rate-limit
        return len(request_timestamps) <= max_requests



import os
import pytest
from aikido_zen.vulnerabilities.sql_injection import (
    detect_sql_injection,
    should_return_early,
)

"""Removed tests :
-> `I'm writting you` : Invalid SQL
-> Moved a lot of the keywords/words together collection to BAD_SQL_COMMANDS.
-> Removed the following GOOD_SQL_COMMANDS : "abcdefghijklmnop@hotmail.com", "steve@yahoo.com"
    Reason : This should never occur unencapsulated in query, results in 5 tokens or so being morphed into one.
"""
BAD_SQL_COMMANDS = [
    "Roses are red insErt are blue",
    "Roses are red cREATE are blue",
    "Roses are red drop are blue",
    "Roses are red updatE are blue",
    "Roses are red SELECT are blue",
    "Roses are red dataBASE are blue",
    "Roses are red alter are blue",
    "Roses are red grant are blue",
    "Roses are red savepoint are blue",
    "Roses are red commit are blue",
    "Roses are red or blue",
    "Roses are red and lovely",
    "This is a group_concat_test",
    "Termin;ate",
    "Roses <> violets",
    "Roses < Violets",
    "Roses > Violets",
    "Roses != Violets",
    "Roses asks red truncates asks blue",
    "Roses asks reddelete asks blue",
    "Roses asks red WHEREis blue",
    "Roses asks red ORis isAND",
    "I was benchmark ing",
    "We were delay ed",
    "I will waitfor you",
]

GOOD_SQL_COMMANDS = [
    "              ",
    "#",
    "'",
]

# Moved ["'union'  is not UNION", "UNION"], to IS_NOT_INJECTION : This is in fact, not an injection.
IS_NOT_INJECTION = [
    ["'UNION 123' UNION \"UNION 123\"", "UNION 123"],
    ["'union'  is not \"UNION\"", "UNION!"],
    ['"UNION;"', "UNION;"],
    ["SELECT * FROM table", "*"],
    ['"COPY/*"', "COPY/*"],
    ["'union'  is not \"UNION--\"", "UNION--"],
    ["'union'  is not UNION", "UNION"],
]

IS_INJECTION = [
    ["UNTER;", "UNTER;"],
]

DIALECTS = [
    "generic",
    "mysql",
    "postgres",
    "sqlite",
]


def is_sql_injection(sql, input, dialect="all"):
    for current in DIALECTS:
        if dialect == "all" or dialect == current:
            result = detect_sql_injection(sql, input, current)
            assert (
                result == True
            ), f"Expected SQL injection for SQL: {sql} and input: {input} in {current} dialect"


def is_not_sql_injection(sql, input, dialect="all"):
    for current in DIALECTS:
        if dialect == "all" or dialect == current:
            result = detect_sql_injection(sql, input, current)
            assert (
                result == False
            ), f"Expected no SQL injection for SQL: {sql} and input: {input} in {current} dialect"


def test_should_return_early():
    # Test cases where the function should return True

    # User input is empty
    assert should_return_early("SELECT * FROM users", "") == True

    # User input is a single character
    assert should_return_early("SELECT * FROM users", "a") == True

    # User input is larger than query
    assert (
        should_return_early("SELECT * FROM users", "SELECT * FROM users WHERE id = 1")
        == True
    )

    # User input not in query
    assert should_return_early("SELECT * FROM users", "DELETE") == True

    # User input is alphanumerical
    assert should_return_early("SELECT * FROM users123", "users123") == True
    assert should_return_early("SELECT * FROM users_123", "users_123") == True
    assert should_return_early("SELECT __1 FROM users_123", "__1") == True
    assert (
        should_return_early(
            "SELECT * FROM table_name_is_fun_12", "table_name_is_fun_12"
        )
        == True
    )

    # User input is a valid comma-separated number list
    assert should_return_early("SELECT * FROM users", "1,2,3") == True

    # User input is a valid number
    assert should_return_early("SELECT * FROM users", "123") == True

    # User input is a valid number with spaces
    assert should_return_early("SELECT * FROM users", "  123  ") == True

    # User input is a valid number with commas
    assert should_return_early("SELECT * FROM users", "1, 2, 3") == True

    # Test cases where the function should return False

    # User input is in query
    assert should_return_early("SELECT * FROM users", " users") == False

    # User input is a valid string in query
    assert should_return_early("SELECT * FROM users", "SELECT ") == False

    # User input is a valid string in query with special characters
    assert (
        should_return_early("SELECT * FROM users; DROP TABLE", "users; DROP TABLE")
        == False
    )


def test_bad_sql_commands():
    for sql in BAD_SQL_COMMANDS:
        is_sql_injection(sql, sql)


def test_good_sql_commands():
    for sql in GOOD_SQL_COMMANDS:
        is_not_sql_injection(sql, sql)


def test_is_injection():
    for sql, input in IS_INJECTION:
        is_sql_injection(sql, input)


def test_is_not_injection():
    for sql, input in IS_NOT_INJECTION:
        is_not_sql_injection(sql, input)


"""
Moved :
is_sql_injection("SELECT * FROM users WHERE id = 'users\\'", "users\\")
is_sql_injection("SELECT * FROM users WHERE id = 'users\\\\'", "users\\\\")
to is_not_sql_injection. Reason : Invalid SQL.
"""


def test_allow_escape_sequences():
    # Invalid queries :
    is_not_sql_injection("SELECT * FROM users WHERE id = 'users\\'", "users\\")

    is_not_sql_injection("SELECT * FROM users WHERE id = 'users\\\\'", "users\\\\")
    is_not_sql_injection("SELECT * FROM users WHERE id = '\nusers'", "\nusers")
    is_not_sql_injection("SELECT * FROM users WHERE id = '\rusers'", "\rusers")
    is_not_sql_injection("SELECT * FROM users WHERE id = '\tusers'", "\tusers")


# Marked "SELECT * FROM users WHERE id IN ('123')", "'123'" as not a sql injection, Reason :
# We replace '123' token with another token and the token count remains the same. This is also
# Not an actual SQL Injection so the algorithm is valid in it's reasoning.
def test_user_input_inside_in():
    is_not_sql_injection("SELECT * FROM users WHERE id IN ('123')", "'123'")
    is_not_sql_injection("SELECT * FROM users WHERE id IN (123)", "123")
    is_not_sql_injection("SELECT * FROM users WHERE id IN (123, 456)", "123")
    is_not_sql_injection("SELECT * FROM users WHERE id IN (123, 456)", "456")
    is_not_sql_injection("SELECT * FROM users WHERE id IN ('123')", "123")
    is_not_sql_injection("SELECT * FROM users WHERE id IN (13,14,15)", "13,14,15")
    is_not_sql_injection("SELECT * FROM users WHERE id IN (13, 14, 154)", "13, 14, 154")
    is_sql_injection(
        "SELECT * FROM users WHERE id IN (13, 14, 154) OR (1=1)", "13, 14, 154) OR (1=1"
    )


# Updated tests so the strings terminate and this is valid SQL.
def test_check_string_safely_escaped():
    is_sql_injection(
        'SELECT * FROM comments WHERE comment = "I" "m writting you"',
        'I" "m writting you',
    )
    is_sql_injection("SELECT * FROM `comm`ents``", "`comm`ents")
    is_not_sql_injection(
        'SELECT * FROM comments WHERE comment = "I\'m writting you"', "I'm writting you"
    )
    is_not_sql_injection(
        "SELECT * FROM comments WHERE comment = 'I\"m writting you'", 'I"m writting you'
    )
    is_not_sql_injection(
        'SELECT * FROM comments WHERE comment = "I`m writting you"', "I`m writting you"
    )
    # Invalid query (strings don't terminate)
    is_not_sql_injection(
        "SELECT * FROM comments WHERE comment = 'I'm writting you'", "I'm writting you"
    )
    # Positive example of same query :
    is_sql_injection(
        "SELECT * FROM comments WHERE comment = 'I'm writting you--'",
        "I'm writting you--",
    )
    is_sql_injection(
        "SELECT * FROM comments WHERE comment = 'I'm writting you''",
        "I'm writting you'",
    )
    # Invalid query in postgres, tests fallback :
    is_sql_injection("SELECT * FROM `comm` ents", "`comm` ents", "postgres")

    # MySQL Specific code :
    is_sql_injection("SELECT * FROM `comm` ents", "`comm` ents", "mysql")
    is_not_sql_injection("SELECT * FROM `comm'ents`", "comm'ents", "mysql")


def test_not_flag_select_queries():
    is_not_sql_injection("SELECT * FROM users WHERE id = 1", "SELECT")


def test_not_flag_escaped_hash():
    is_not_sql_injection("SELECT * FROM hashtags WHERE name = '#hashtag'", "#hashtag")


def test_comment_same_as_user_input():
    is_sql_injection(
        "SELECT * FROM hashtags WHERE name = '-- Query by name' -- Query by name",
        "-- Query by name",
    )


def test_input_occurs_in_comment():
    is_not_sql_injection(
        "SELECT * FROM hashtags WHERE name = 'name' -- Query by name", "name"
    )


def test_user_input_is_multiline():
    is_sql_injection(
        "SELECT * FROM users WHERE id = 'a'\nOR 1=1#'", "a'\nOR 1=1#", "mysql"
    )
    is_sql_injection(
        "SELECT * FROM users WHERE id = 'a'\nOR 1=1#'", "a'\nOR 1=1#", "generic"
    )
    is_not_sql_injection("SELECT * FROM users WHERE id = 'a\nb\nc';", "a\nb\nc")


def test_user_input_is_longer_than_query():
    is_not_sql_injection("SELECT * FROM users", "SELECT * FROM users WHERE id = 'a'")


def test_sqlite_dollar_placeholder():
    is_sql_injection(
        "SELECT * FROM users WHERE id = '1' OR $$ IS NULL -- '",
        "1' OR $$ IS NULL -- ",
        "sqlite",
    )


def test_multiline_queries():
    is_sql_injection(
        """
        SELECT * FROM `users`
        WHERE id = 123
        """,
        "users`",
    )

    is_sql_injection(
        """
        SELECT *
        FROM users
        WHERE id = '1' OR 1=1
        """,
        "1' OR 1=1",
    )

    is_sql_injection(
        """
        SELECT *
        FROM users
        WHERE id = '1' OR 1=1
        AND is_escaped = '1'' OR 1=1'
        """,
        "1' OR 1=1",
    )

    is_sql_injection(
        """
        SELECT *
        FROM users
        WHERE id = '1' OR 1=1
        AND is_escaped = "1' OR 1=1"
        """,
        "1' OR 1=1",
    )

    is_not_sql_injection(
        """
        SELECT * FROM `users`
        WHERE id = 123
        """,
        "123",
    )

    is_not_sql_injection(
        """
        SELECT * FROM `us``ers`
        WHERE id = 123
        """,
        "users",
    )

    is_not_sql_injection(
        """
        SELECT * FROM users
        WHERE id = 123
        """,
        "123",
    )

    is_not_sql_injection(
        """
        SELECT * FROM users
        WHERE id = '123'
        """,
        "123",
    )

    is_not_sql_injection(
        """
        SELECT *
        FROM users
        WHERE is_escaped = "1' OR 1=1"
        """,
        "1' OR 1=1",
    )


def test_lowercased_input_sql_injection():
    sql = """
        SELECT id,
               email,
               password_hash,
               registered_at,
               is_confirmed,
               first_name,
               last_name
        FROM users WHERE email_lowercase = '' or 1=1 -- a'
    """
    expected_sql_injection = "' OR 1=1 -- a"

    is_sql_injection(sql, expected_sql_injection)


# Removed dangerous character parametized loop : This was all producing invalid SQL.

"""
Marked the following as SQL injection since this would result in 2 or more tokens becoming one :
is_not_sql_injection("foobar)", "foobar)")
is_not_sql_injection("foobar      )", "foobar      )")
is_not_sql_injection("€foobar()", "€foobar()")
"""


def test_function_calls_as_sql_injections():
    is_sql_injection("foobar()", "foobar()")
    is_sql_injection("foobar(1234567)", "foobar(1234567)")
    is_sql_injection("foobar       ()", "foobar       ()")
    is_sql_injection(".foobar()", ".foobar()")
    is_sql_injection("20+foobar()", "20+foobar()")
    is_sql_injection("20-foobar(", "20-foobar(")
    is_sql_injection("20<foobar()", "20<foobar()")
    is_sql_injection("20*foobar  ()", "20*foobar  ()")
    is_sql_injection("!foobar()", "!foobar()")
    is_sql_injection("=foobar()", "=foobar()")
    is_sql_injection("1foobar()", "1foobar()")
    is_sql_injection("1foo_bar()", "1foo_bar()")
    is_sql_injection("1foo-bar()", "1foo-bar()")
    is_sql_injection("#foobar()", "#foobar()")
    is_sql_injection("foobar)", "foobar)")
    is_sql_injection("foobar      )", "foobar      )")
    is_sql_injection("€foobar()", "€foobar()")


def file_paths():
    script_dir = os.path.dirname(__file__)
    return [
        # Taken from https://github.com/payloadbox/sql-injection-payload-list/tree/master
        os.path.join(script_dir, "payloads/Auth_Bypass.txt"),
        os.path.join(script_dir, "payloads/postgres.txt"),
        os.path.join(script_dir, "payloads/mysql.txt"),
        os.path.join(script_dir, "payloads/mssql_and_db2.txt"),
    ]


# Define the Pytest tests
for file_path in file_paths():
    import MySQLdb

    with open(file_path, "r", encoding="utf-8") as file:
        for line in file:
            sql = line.rstrip("\n")
            escaped_sql = MySQLdb._mysql.escape_string(sql).decode("utf-8")

            def test_sql_injection():
                is_sql_injection(sql, sql)
                is_not_sql_injection(f"'{escaped_sql}'", escaped_sql, "mysql")
                is_not_sql_injection(f"'{escaped_sql}'", sql, "mysql")

            def test_sql_injection_query():
                is_sql_injection(f"SELECT * FROM users WHERE id = {sql}", sql)
                is_not_sql_injection(
                    f"SELECT * FROM users WHERE id = '{escaped_sql}'",
                    escaped_sql,
                    "mysql",
                )
                is_not_sql_injection(
                    f"SELECT * FROM users WHERE id = '{escaped_sql}'", sql, "mysql"
                )



import pytest
from aikido_zen.helpers.get_subdomains_from_url import get_subdomains_from_url


def test_get_subdomains_from_url():
    # Test cases with expected results
    test_cases = [
        # Test with a standard URL
        ("http://tobi.ferrets.example.com", ["tobi", "ferrets"]),
        # Test with a URL that has no subdomains
        ("http://example.com", []),
        # Test with a URL that has multiple subdomains
        ("http://a.b.c.example.com", ["a", "b", "c"]),
        # Test with a URL that has a port
        ("http://tobi.ferrets.example.com:8080", ["tobi", "ferrets"]),
        # Test with a URL that has only the main domain
        ("http://localhost", []),
        # Test with an invalid URL
        ("http://.com", []),
        # Test with an empty string
        ("", []),
        # Test with a URL with subdomains and a path
        ("http://tobi.ferrets.example.com/path/to/resource", ["tobi", "ferrets"]),
        ({}, []),
        (None, []),
    ]

    for url, expected in test_cases:
        assert get_subdomains_from_url(url) == expected



"""
Code to poll for realtime changes
"""

import os
import requests
from aikido_zen.helpers.logging import logger
from aikido_zen.helpers.urls.get_api_url import get_api_url


def get_realtime_url():
    """Fetches the default realtime url or environment variable"""
    realtime_url = os.getenv("AIKIDO_REALTIME_ENDPOINT")
    if realtime_url is not None:
        if not realtime_url.endswith("/"):
            realtime_url += "/"
        return realtime_url
    return "https://runtime.aikido.dev/"


def get_config(token):
    """Fetches the config from realtime URL"""
    url = get_api_url() + "api/runtime/config"
    headers = {
        "Authorization": str(token),
    }
    response = requests.get(url, headers=headers, timeout=3)  # timeout in 3 seconds
    if response.status_code != 200:
        logger.info("Invalid response from api : %s", response.status_code)

    return response.json()  # Parse and return the JSON response


def get_config_last_updated_at(token):
    """
    Fetches the time when the config was last updated from realtime server
    """
    url = f"{get_realtime_url()}config"
    headers = {
        "Authorization": str(token),
    }
    response = requests.get(url, headers=headers, timeout=0.5)  # timeout in 500ms
    if response.status_code != 200:
        logger.info("Invalid response from realtime api : %s", response.status_code)

    return int(
        response.json().get("configUpdatedAt", 0)
    )  #  Return configUpdatedAt time



import pytest
from .merge_data_schemas import merge_data_schemas
from .get_api_info import get_data_schema


@pytest.fixture
def schema_a():
    return {
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "age": {"type": "integer"},
        },
    }


@pytest.fixture
def schema_b():
    return {
        "type": "object",
        "properties": {
            "age": {"type": "integer"},
            "email": {"type": "string"},
        },
    }


@pytest.fixture
def schema_c():
    return {"type": "null"}


@pytest.fixture
def schema_d():
    return {
        "type": "object",
        "properties": {
            "address": {"type": "string"},
        },
    }


@pytest.fixture
def schema_e():
    return {
        "type": "array",
        "items": {"type": "string"},
    }


@pytest.fixture
def schema_f():
    return {"type": "array", "items": {"type": "integer"}}


def test_merge_same_type_schemas(schema_a, schema_b):
    merged = merge_data_schemas(schema_a, schema_b)
    props = merged.get("properties")
    assert props["name"].get("type") == "string"
    assert props["age"].get("type") == "integer"
    assert props["email"].get("type") == "string"
    assert props["age"].get("optional") is None
    assert props["email"].get("optional") is True


def test_merge_different_type_schemas(schema_a, schema_c):
    merged = merge_data_schemas(schema_a, schema_c)
    assert merged == schema_a


def test_merge_null_type_preference(schema_c, schema_b):
    merged = merge_data_schemas(schema_c, schema_b)
    assert merged == schema_b


def test_merge_with_optional_properties(schema_a, schema_d):
    merged = merge_data_schemas(schema_a, schema_d)
    assert "address" in merged["properties"]
    assert merged["properties"]["address"]["optional"] is True


def test_merge_empty_properties(schema_a):
    empty_schema = {"type": "object", "properties": {}}
    merged = merge_data_schemas(schema_a, empty_schema)
    assert merged["properties"] == schema_a["properties"]


def test_merge_with_no_properties(schema_a):
    no_properties_schema = {"type": "object", "properties": {}}
    merged = merge_data_schemas(no_properties_schema, schema_a)
    assert merged["properties"] == schema_a["properties"]


def test_merge_with_nested_schemas(schema_a, schema_b):
    nested_schema_a = {"type": "object", "properties": {"details": schema_a}}
    nested_schema_b = {"type": "object", "properties": {"details": schema_b}}
    merged = merge_data_schemas(nested_schema_a, nested_schema_b)
    assert "details" in merged["properties"]
    assert merged["properties"]["details"]["properties"]["name"]["type"] == "string"
    assert merged["properties"]["details"]["properties"]["email"]["type"] == "string"
    assert merged["properties"]["details"]["properties"]["age"]["type"] is "integer"


def test_merge_with_no_items(schema_e):
    empty_array_schema = {"type": "array", "items": None}
    merged = merge_data_schemas(schema_e, empty_array_schema)
    assert merged["items"] == schema_e["items"]


def test_merge_with_different_item_types(schema_e, schema_f):
    merged = merge_data_schemas(schema_e, schema_f)
    assert set(merged["items"].get("type")) == {"string", "integer"}


def test_empty_array():
    assert merge_data_schemas(get_data_schema([]), get_data_schema([])) == {
        "type": "array",
        "items": None,
    }


def test_merge_types():
    assert set(
        merge_data_schemas(get_data_schema("str"), get_data_schema(15))["type"]
    ) == {"string", "number"}

    # Cannot merge object with primitive type
    assert merge_data_schemas(
        get_data_schema({"test": "abc"}), get_data_schema(15)
    ) == {
        "type": "object",
        "properties": {
            "test": {
                "type": "string",
            },
        },
    }

    merged_data_schema1 = merge_data_schemas(
        get_data_schema({"test": "abc"}), get_data_schema({"test": True})
    )
    merged_data_schema1["properties"]["test"]["type"] = set(
        merged_data_schema1["properties"]["test"]["type"]
    )
    assert merged_data_schema1 == {
        "type": "object",
        "properties": {
            "test": {
                "type": {"boolean", "string"},
            },
        },
    }

    merged_data_schema2 = merge_data_schemas(
        get_data_schema({"test": "abc"}),
        merge_data_schemas(
            get_data_schema({"test": "abc"}), get_data_schema({"test": True})
        ),
    )
    merged_data_schema2["properties"]["test"]["type"] = set(
        merged_data_schema2["properties"]["test"]["type"]
    )
    assert merged_data_schema2 == {
        "type": "object",
        "properties": {
            "test": {
                "type": {"boolean", "string"},
            },
        },
    }

    merged_data_schema3 = merge_data_schemas(
        merge_data_schemas(
            get_data_schema({"test": True}), get_data_schema({"test": "test"})
        ),
        get_data_schema({"test": "abc"}),
    )
    merged_data_schema3["properties"]["test"]["type"] = set(
        merged_data_schema3["properties"]["test"]["type"]
    )
    assert merged_data_schema3 == {
        "type": "object",
        "properties": {
            "test": {
                "type": {"string", "boolean"},
            },
        },
    }

    merged_data_schema4 = merge_data_schemas(
        merge_data_schemas(
            get_data_schema({"test": "abc"}), get_data_schema({"test": 123})
        ),
        get_data_schema({"test": True}),
    )
    merged_data_schema4["properties"]["test"]["type"] = set(
        merged_data_schema4["properties"]["test"]["type"]
    )
    assert merged_data_schema4 == {
        "type": "object",
        "properties": {
            "test": {
                "type": {"number", "boolean", "string"},
            },
        },
    }

    merged_data_schema5 = merge_data_schemas(
        merge_data_schemas(
            get_data_schema({"test": "test"}), get_data_schema({"test": True})
        ),
        merge_data_schemas(
            get_data_schema({"test": 123}), get_data_schema({"test": True})
        ),
    )
    merged_data_schema5["properties"]["test"]["type"] = set(
        merged_data_schema5["properties"]["test"]["type"]
    )
    assert merged_data_schema5 == {
        "type": "object",
        "properties": {
            "test": {
                "type": {"string", "boolean", "number"},
            },
        },
    }



from wrapt import wrap_object, FunctionWrapper, when_imported
from aikido_zen.background_process.packages import ANY_VERSION, is_package_compatible
from aikido_zen.errors import AikidoException
from aikido_zen.helpers.logging import logger


def on_import(name, package="", version_requirement=ANY_VERSION):
    """
    Decorator to register a function to be called when a package is imported.
    It checks if the package is compatible with the specified version requirement.
    """

    def decorator(func):
        def check_pkg_wrapper(f):
            def wrapper(*args, **kwargs):
                # This code runs only on import
                if package and not is_package_compatible(package, version_requirement):
                    return
                return f(*args, **kwargs)

            return wrapper

        # Register the function to be called on import
        when_imported(name)(check_pkg_wrapper(func))

    return decorator


def patch_function(module, name, wrapper):
    """
    Patches a function in the specified module with a wrapper function.
    """
    try:
        wrap_object(module, name, FunctionWrapper, (wrapper,))
    except Exception as e:
        logger.info("Failed to wrap %s:%s, due to: %s", module, name, e)


def before(wrapper):
    """
    Surrounds a patch with try-except and calls the original function at the end
    """

    def decorator(func, instance, args, kwargs):
        try:
            wrapper(func, instance, args, kwargs)  # Call the patch
        except AikidoException as e:
            raise e  # Re-raise AikidoException
        except Exception as e:
            logger.debug(
                "%s:%s wrapping-before error: %s", func.__module__, func.__name__, e
            )
        return func(*args, **kwargs)  # Call the original function

    return decorator


def before_async(wrapper):
    """
    Surrounds an async patch with try-except and calls the original asynchronous function at the end
    """

    async def decorator(func, instance, args, kwargs):
        try:
            await wrapper(func, instance, args, kwargs)  # Call the patch
        except AikidoException as e:
            raise e  # Re-raise AikidoException
        except Exception as e:
            logger.debug(
                "%s:%s wrapping-before error: %s", func.__module__, func.__name__, e
            )
        return await func(*args, **kwargs)  # Call the original function

    return decorator


def before_modify_return(wrapper):
    """
    Surrounds a patch with try-except and calls the original function at the end unless a return value is present.
    """

    def decorator(func, instance, args, kwargs):
        try:
            rv = wrapper(func, instance, args, kwargs)  # Call the patch
            if rv is not None:
                return rv
        except AikidoException as e:
            raise e  # Re-raise AikidoException
        except Exception as e:
            logger.debug(
                "%s:%s wrapping-before error: %s", func.__module__, func.__name__, e
            )
        return func(*args, **kwargs)  # Call the original function

    return decorator


def after(wrapper):
    """
    Surrounds a patch with try-except, calls the original function and gives the return value to the patch
    """

    def decorator(func, instance, args, kwargs):
        return_value = func(*args, **kwargs)  # Call the original function
        try:
            wrapper(func, instance, args, kwargs, return_value)  # Call the patch
        except AikidoException as e:
            raise e  # Re-raise AikidoException
        except Exception as e:
            logger.debug(
                "%s:%s wrapping-after error: %s", func.__module__, func.__name__, e
            )

        return return_value

    return decorator



import time

import pytest

from . import current_context, Context
from .users import validate_user, set_user
from .. import should_block_request
from ..thread.thread_cache import get_cache


@pytest.fixture(autouse=True)
def run_around_tests():
    get_cache().reset()
    yield
    # Make sure to reset context and cache after every test so it does not
    # interfere with other tests
    current_context.set(None)
    get_cache().reset()


def set_context_and_lifecycle():
    wsgi_request = {
        "REQUEST_METHOD": "GET",
        "HTTP_HEADER_1": "header 1 value",
        "HTTP_HEADER_2": "Header 2 value",
        "RANDOM_VALUE": "Random value",
        "HTTP_COOKIE": "sessionId=abc123xyz456;",
        "wsgi.url_scheme": "http",
        "HTTP_HOST": "localhost:8080",
        "PATH_INFO": "/hello",
        "QUERY_STRING": "user=JohnDoe&age=30&age=35",
        "CONTENT_TYPE": "application/json",
        "REMOTE_ADDR": "198.51.100.23",
    }
    context = Context(
        req=wsgi_request,
        body=None,
        source="flask",
    )
    context.set_as_current_context()
    return context


def test_validate_user_valid_input():
    user = {"id": "123", "name": "Alice"}
    result = validate_user(user)
    assert result == {"id": "123", "name": "Alice"}


def test_validate_user_valid_input_with_int_id():
    user = {"id": 456, "name": "Bob"}
    result = validate_user(user)
    assert result == {"id": "456", "name": "Bob"}


def test_validate_user_missing_id(caplog):
    user = {"name": "Charlie"}
    result = validate_user(user)
    assert result is None
    assert "expects an object with 'id' property." in caplog.text


def test_validate_user_invalid_id_type(caplog):
    user = {"id": 12.34, "name": "David"}
    result = validate_user(user)
    assert result is None
    assert (
        "expects an object with 'id' property of type string or number" in caplog.text
    )


def test_validate_user_empty_string_id(caplog):
    user = {"id": "", "name": "Eve"}
    result = validate_user(user)
    assert result is None
    assert "expects an object with 'id' property non-empty string." in caplog.text


def test_validate_user_missing_name(caplog):
    user = {"id": "789"}
    result = validate_user(user)
    assert result == {"id": "789"}


def test_validate_user_empty_name(caplog):
    user = {"id": "101", "name": ""}
    result = validate_user(user)
    assert result == {"id": "101"}


def test_validate_user_invalid_user_type(caplog):
    user = ["id", "name"]
    result = validate_user(user)
    assert result is None
    assert "expects a dict with 'id' and 'name' properties" in caplog.text


def test_validate_user_invalid_user_type_dict_without_id(caplog):
    user = {"name": "Frank"}
    result = validate_user(user)
    assert result is None
    assert "expects an object with 'id' property." in caplog.text


def test_set_user_with_none(caplog):
    result = set_user(None)
    assert "expects a dict with 'id' and 'name' properties" in caplog.text


def test_set_valid_user():
    context1 = set_context_and_lifecycle()
    assert context1.user is None

    user = {"id": 456, "name": "Bob"}
    set_user(user)

    assert context1.user == {
        "id": "456",
        "name": "Bob",
        "lastIpAddress": "198.51.100.23",
    }

    assert len(get_cache().users.as_array()) == 1
    user_1 = get_cache().users.as_array()[0]
    assert user_1["id"] == "456"
    assert user_1["lastIpAddress"] == "198.51.100.23"
    assert user_1["name"] == "Bob"
    assert user_1["firstSeenAt"] == user_1["lastSeenAt"]


def test_re_set_valid_user():
    context1 = set_context_and_lifecycle()
    assert context1.user is None

    user = {"id": 456, "name": "Bob"}
    set_user(user)

    assert context1.user == {
        "id": "456",
        "name": "Bob",
        "lastIpAddress": "198.51.100.23",
    }
    assert len(get_cache().users.as_array()) == 1
    user_1 = get_cache().users.as_array()[0]
    assert user_1["id"] == "456"
    assert user_1["lastIpAddress"] == "198.51.100.23"
    assert user_1["name"] == "Bob"
    assert user_1["firstSeenAt"] == user_1["lastSeenAt"]

    user = {"id": "1000", "name": "Alice"}
    set_user(user)

    assert context1.user == {
        "id": "1000",
        "name": "Alice",
        "lastIpAddress": "198.51.100.23",
    }

    assert len(get_cache().users.as_array()) == 2
    assert get_cache().users.as_array()[0] == user_1
    user_2 = get_cache().users.as_array()[1]
    assert user_2["id"] == "1000"
    assert user_2["lastIpAddress"] == "198.51.100.23"
    assert user_2["name"] == "Alice"


def test_after_middleware(caplog):
    context1 = set_context_and_lifecycle()
    assert context1.user is None

    should_block_request()

    user = {"id": 456, "name": "Bob"}
    set_user(user)

    assert "must be called before the Zen middleware is executed" in caplog.text
    assert context1.user == {
        "id": "456",
        "name": "Bob",
        "lastIpAddress": "198.51.100.23",
    }


def test_set_valid_user_twice():
    context1 = set_context_and_lifecycle()
    assert context1.user is None

    user = {"id": 456, "name": "Bob"}
    set_user(user)

    assert context1.user == {
        "id": "456",
        "name": "Bob",
        "lastIpAddress": "198.51.100.23",
    }
    assert len(get_cache().users.as_array()) == 1
    first_seen_at = int(get_cache().users.as_array()[0]["firstSeenAt"])

    time.sleep(1)
    set_user(user)  # 2nd time

    assert len(get_cache().users.as_array()) == 1
    user_1 = get_cache().users.as_array()[0]
    assert user_1["id"] == "456"
    assert user_1["lastIpAddress"] == "198.51.100.23"
    assert user_1["name"] == "Bob"
    assert user_1["firstSeenAt"] != user_1["lastSeenAt"]
    assert user_1["firstSeenAt"] == first_seen_at



"""
Mainly exports `start_polling_for_changes` function
"""

from aikido_zen.helpers.token import Token
from aikido_zen.helpers.logging import logger
import aikido_zen.background_process.realtime as realtime

POLL_FOR_CONFIG_CHANGES_INTERVAL = 60  #  Poll for config changes every 60 seconds


def start_polling_for_changes(connection_manager, event_scheduler):
    """
    Arguments :
    - on_config_update : A function that will run with the new config if changed
    - serverless, token : Attributes from the connection_manager
    - last_updated_at : The last time the config was updated (unixtime in ms)
    This function will check if the config was updated or not
    """
    if not isinstance(connection_manager.token, Token):
        logger.info("No token provided, not polling for config updates")
        return
    if connection_manager.serverless:
        logger.info("Running in serverless environment, not polling for config updates")
        return

    # Start the interval by booting the first settimeout
    poll_for_changes(
        connection_manager=connection_manager,
        token=connection_manager.token,
        former_last_updated=connection_manager.conf.last_updated_at,
        event_scheduler=event_scheduler,
    )


def poll_for_changes(
    connection_manager,
    token,
    former_last_updated,
    event_scheduler,
):
    """
    Actually performs the check if the config was updated or not
    """
    # If something went wrong, or we don't know when the config was
    # last updated, set to prev value
    last_updated = former_last_updated
    try:
        last_updated = realtime.get_config_last_updated_at(token)
        config_changed = (
            isinstance(former_last_updated, int) and last_updated > former_last_updated
        )
        if config_changed:
            #  The config changed
            logger.debug("According to realtime: Config changed")
            config = realtime.get_config(token)
            connection_manager.update_service_config({**config, "success": True})
            connection_manager.update_firewall_lists()
    except Exception as e:
        logger.error("Failed to check for config updates due to error : %s", e)

    # Set a timeout for `POLL_FOR_CONFIG_CHANGES_INTERVAL` seconds until this function
    # Gets called again
    event_scheduler.enter(
        POLL_FOR_CONFIG_CHANGES_INTERVAL,
        1,
        poll_for_changes,
        (connection_manager, token, last_updated, event_scheduler),
    )



import pytest
from .hostnames import Hostnames


def test_add_hostname():
    """Test adding a hostname."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", 80)
    key = "example.com:80"
    assert key in hostnames.map
    assert hostnames.map[key]["hostname"] == "example.com"
    assert hostnames.map[key]["port"] == 80
    assert hostnames.map[key]["hits"] == 1


def test_add_multiple_ports():
    """Test adding multiple ports for the same hostname."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", 80)
    hostnames.add("example.com", 443)

    key80 = "example.com:80"
    key443 = "example.com:443"

    assert key80 in hostnames.map
    assert key443 in hostnames.map
    assert hostnames.map[key80]["hits"] == 1
    assert hostnames.map[key443]["hits"] == 1


def test_add_duplicate_hostname():
    """Test adding a duplicate hostname."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", 80)
    hostnames.add("example.com", 80)  # Should not change the port
    key = "example.com:80"
    assert hostnames.map[key]["hits"] == 2  # Hits should increment


def test_max_entries():
    """Test that the maximum number of entries is enforced."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", 80)
    hostnames.add("test.com", 443)
    hostnames.add("localhost", None)
    hostnames.add("newsite.com", 8080)  # This should remove "example.com"

    assert "example.com:80" not in hostnames.map
    assert "test.com:443" in hostnames.map
    assert "localhost:None" in hostnames.map
    assert "newsite.com:8080" in hostnames.map
    assert len(hostnames.map) == 3


def test_as_array():
    """Test the as_array method."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", 80)
    hostnames.add("test.com", 443)
    expected_array = [
        {"hostname": "example.com", "port": 80, "hits": 1},
        {"hostname": "test.com", "port": 443, "hits": 1},
    ]
    assert hostnames.as_array() == expected_array
    assert isinstance(hostnames.as_array(), list)


def test_hostnames_add_multiple():
    """Test the as_array method."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", 80, 20)
    hostnames.add("test.com", 443)
    hostnames.add("test.com", 443, 4)
    expected_array = [
        {"hostname": "example.com", "port": 80, "hits": 20},
        {"hostname": "test.com", "port": 443, "hits": 5},
    ]
    assert hostnames.as_array() == expected_array
    assert isinstance(hostnames.as_array(), list)


def test_clear():
    """Test the clear method."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", 80)
    hostnames.add("test.com", 443)
    hostnames.clear()
    assert len(hostnames.map) == 0


def test_add_none_port():
    """Test adding a hostname with a None port."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", None)
    key = "example.com:None"
    assert key in hostnames.map
    assert hostnames.map[key]["hostname"] == "example.com"
    assert hostnames.map[key]["port"] is None
    assert hostnames.map[key]["hits"] == 1


def test_exceed_max_entries_with_multiple_ports():
    """Test exceeding max entries with multiple ports."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", 80)
    hostnames.add("example.com", 443)
    hostnames.add("test.com", 8080)
    hostnames.add("newsite.com", 3000)  # This should remove "example.com:80"

    assert "example.com:80" not in hostnames.map
    assert "example.com:443" in hostnames.map
    assert "test.com:8080" in hostnames.map
    assert "newsite.com:3000" in hostnames.map
    assert len(hostnames.map) == 3


def test_add_and_remove_ports():
    """Test adding and removing ports."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", 80)
    assert hostnames.map["example.com:80"]["hits"] == 1
    hostnames.add("example.com", 443)
    assert hostnames.map["example.com:80"]["hits"] == 1
    assert hostnames.map["example.com:443"]["hits"] == 1
    hostnames.add("test.com", 8080)
    assert hostnames.map["example.com:80"]["hits"] == 1
    assert hostnames.map["example.com:443"]["hits"] == 1
    assert hostnames.map["test.com:8080"]["hits"] == 1

    # Remove a port (not directly supported in the current implementation)
    # This part of the test is not applicable since we don't have a remove method.
    # Instead, we can just check the hits.
    hostnames.map["example.com:80"]["hits"] = 0  # Simulating removal
    assert hostnames.map["example.com:80"]["hits"] == 0  # Should only have port 443


def test_clear_with_multiple_entries():
    """Test clearing with multiple entries."""
    hostnames = Hostnames(max_entries=3)
    hostnames.add("example.com", 80)
    hostnames.add("test.com", 443)
    hostnames.add("localhost", 3000)
    hostnames.clear()
    assert len(hostnames.map) == 0



"""Exports AikidoDjangoMiddleware"""

from aikido_zen.helpers.logging import logger
from .should_block_request import should_block_request


class AikidoDjangoMiddleware:
    """Middleware for rate-limiting and user blocking for django"""

    def __init__(self, get_response):
        logger.critical("Django middleware ised")
        self.get_response = get_response
        try:
            from django.http import HttpResponse

            self.HttpResponse = HttpResponse
        except ImportError:
            logger.warning(
                "django.http import not working, aikido rate-limiting middleware not running."
            )

    def __call__(self, request):
        result = should_block_request()
        if result["block"] is not True or self.HttpResponse is None:
            return self.get_response(request)

        if result["type"] == "ratelimited":
            message = "You are rate limited by Zen."
            if result["trigger"] == "ip" and result["ip"]:
                message += " (Your IP: " + result["ip"] + ")"
            return self.HttpResponse(message, content_type="text/plain", status=429)
        elif result["type"] == "blocked":
            return self.HttpResponse(
                "You are blocked by Zen.", content_type="text/plain", status=403
            )
        logger.debug("Unknown type for blocking request: %s", result["type"])
        return self.get_response(request)



import aikido_zen.sinks.clickhouse_driver
import pytest
from aikido_zen.background_process import reset_comms
from aikido_zen.context import Context
from aikido_zen.errors import AikidoSQLInjection


class Context1(Context):
    def __init__(self, body):
        self.cookies = {}
        self.headers = {}
        self.remote_address = "1.1.1.1"
        self.method = "POST"
        self.url = "url"
        self.query = {}
        self.body = body
        self.source = "express"
        self.route = "/"
        self.parsed_userinput = {}


@pytest.fixture(autouse=True)
def set_blocking_to_true(monkeypatch):
    monkeypatch.setenv("AIKIDO_BLOCK", "1")


@pytest.fixture
def client():
    from clickhouse_driver import Client

    return Client(
        host="127.0.0.1", port=9000, user="default", password="", database="default"
    )


def test_client_execute_without_context(client):
    reset_comms()
    dog_name = "Steve"
    sql = "INSERT INTO dogs (dog_name, isAdmin) VALUES ('{}' , 0)".format(dog_name)
    client.execute(sql)


def test_client_execute_safe(client):
    reset_comms()
    dog_name = "Steve"
    sql = "INSERT INTO dogs (dog_name, isAdmin) VALUES ('{}' , 0)".format(dog_name)
    Context1({"dog_name": dog_name}).set_as_current_context()
    client.execute(sql)


def test_client_execute_unsafe(client, monkeypatch):
    reset_comms()
    dog_name = "Malicious dog', 1); -- "
    sql = "INSERT INTO dogs (dog_name, isAdmin) VALUES ('{}' , 0)".format(dog_name)
    Context1({"dog_name": dog_name}).set_as_current_context()

    with pytest.raises(AikidoSQLInjection):
        client.execute(sql)

    monkeypatch.setenv("AIKIDO_BLOCK", "0")
    client.execute(sql)


def test_cursor_execute_safe():
    from clickhouse_driver import connect

    conn = connect("clickhouse://localhost:9000")
    reset_comms()
    dog_name = "Steve"
    sql = "INSERT INTO dogs (dog_name, isAdmin) VALUES ('{}' , 0)".format(dog_name)
    Context1({"dog_name": dog_name}).set_as_current_context()
    conn.cursor().execute(sql)


def test_cursor_execute_unsafe(monkeypatch):
    from clickhouse_driver import connect

    conn = connect("clickhouse://localhost:9000")
    reset_comms()
    dog_name = "Malicious dog', 1); -- "
    sql = "INSERT INTO dogs (dog_name, isAdmin) VALUES ('{}' , 0)".format(dog_name)
    Context1({"dog_name": dog_name}).set_as_current_context()

    with pytest.raises(AikidoSQLInjection):
        conn.cursor().execute(sql)

    monkeypatch.setenv("AIKIDO_BLOCK", "0")
    conn.cursor().execute(sql)


def test_client_execute_with_progress_safe(client):
    reset_comms()
    dog_name = "Steve"
    sql = "INSERT INTO dogs (dog_name, isAdmin) VALUES ('{}' , 0)".format(dog_name)
    Context1({"dog_name": dog_name}).set_as_current_context()
    client.execute_with_progress(sql)


def test_client_execute_with_progress_unsafe(client, monkeypatch):
    reset_comms()
    dog_name = "Malicious dog', 1); -- "
    sql = "INSERT INTO dogs (dog_name, isAdmin) VALUES ('{}' , 0)".format(dog_name)
    Context1({"dog_name": dog_name}).set_as_current_context()

    with pytest.raises(AikidoSQLInjection):
        client.execute_with_progress(sql)

    monkeypatch.setenv("AIKIDO_BLOCK", "0")
    client.execute_with_progress(sql)


def test_client_execute_iter_safe(client):
    reset_comms()
    dog_name = "Steve"
    sql = "INSERT INTO dogs (dog_name, isAdmin) VALUES ('{}' , 0)".format(dog_name)
    Context1({"dog_name": dog_name}).set_as_current_context()
    client.execute_iter(sql)


def test_client_execute_iter_unsafe(client, monkeypatch):
    reset_comms()
    dog_name = "Malicious dog', 1); -- "
    sql = "INSERT INTO dogs (dog_name, isAdmin) VALUES ('{}' , 0)".format(dog_name)
    Context1({"dog_name": dog_name}).set_as_current_context()

    with pytest.raises(AikidoSQLInjection):
        client.execute_iter(sql)

    monkeypatch.setenv("AIKIDO_BLOCK", "0")
    client.execute_iter(sql)



"""
init file for the helpers/ folder, imports helper function for easier accessibility
"""

from aikido_zen.helpers.escape_string_regexp import escape_string_regexp
from aikido_zen.helpers.get_current_and_next_segments import (
    get_current_and_next_segments,
)



import pytest
from . import Statistics, Operations


@pytest.fixture
def stats():
    return Statistics()


def test_initialization(monkeypatch):
    # Mock the current time
    mock_time = 1234567890000
    monkeypatch.setattr(
        "aikido_zen.helpers.get_current_unixtime_ms.get_unixtime_ms", lambda: mock_time
    )

    stats = Statistics()
    assert stats.total_hits == 0
    assert stats.attacks_detected == 0
    assert stats.attacks_blocked == 0
    assert stats.started_at == mock_time
    assert isinstance(stats.operations, Operations)


def test_clear(monkeypatch):
    # Mock the current time
    mock_time = 1234567890000
    monkeypatch.setattr(
        "aikido_zen.helpers.get_current_unixtime_ms.get_unixtime_ms", lambda: mock_time
    )

    stats = Statistics()
    stats.total_hits = 10
    stats.attacks_detected = 5
    stats.attacks_blocked = 3
    stats.operations.register_call("test", "sql_op")
    stats.clear()

    assert stats.total_hits == 0
    assert stats.attacks_detected == 0
    assert stats.attacks_blocked == 0
    assert stats.started_at == mock_time
    assert stats.operations == {}


def test_increment_total_hits():
    stats = Statistics()
    stats.increment_total_hits()
    assert stats.total_hits == 1


def test_on_detected_attack(stats):
    stats.on_detected_attack(blocked=True, operation="test_op")
    assert stats.attacks_detected == 1
    assert stats.attacks_blocked == 1

    stats.on_detected_attack(blocked=False, operation="test_op")
    assert stats.attacks_detected == 2
    assert stats.attacks_blocked == 1


def test_get_record(monkeypatch):
    # Mock the current time
    mock_time = 1234567890000
    monkeypatch.setattr(
        "aikido_zen.helpers.get_current_unixtime_ms.get_unixtime_ms", lambda: mock_time
    )

    stats = Statistics()
    stats.total_hits = 10
    stats.operations.register_call("test.test", "nosql_op")
    stats.on_detected_attack(blocked=True, operation="test.test")
    stats.attacks_detected = 5
    stats.attacks_blocked = 3

    record = stats.get_record()
    assert record["startedAt"] == stats.started_at
    assert record["endedAt"] == mock_time
    assert record["requests"]["total"] == 10
    assert record["requests"]["aborted"] == 0
    assert record["requests"]["attacksDetected"]["total"] == 5
    assert record["requests"]["attacksDetected"]["blocked"] == 3
    assert record["operations"] == {
        "test.test": {
            "attacksDetected": {"blocked": 1, "total": 1},
            "kind": "nosql_op",
            "total": 1,
        }
    }


def test_import_from_record():
    stats = Statistics()
    stats.operations.register_call("test.test", "nosql_op")
    stats.operations.register_call("test.test", "nosql_op")
    stats.operations.register_call("test.test", "nosql_op")
    record = {
        "requests": {
            "total": 10,
            "attacksDetected": {
                "total": 5,
                "blocked": 3,
            },
        },
        "operations": {
            "test.test": {
                "attacksDetected": {"blocked": 1, "total": 1},
                "kind": "nosql_op",
                "total": 1,
            },
            "test.test2": {
                "kind": "sql_op",
                "total": 5,
                "attacksDetected": {"blocked": 5, "total": 200},
            },
        },
    }
    stats.import_from_record(record)
    assert stats.total_hits == 10
    assert stats.attacks_detected == 5
    assert stats.attacks_blocked == 3
    assert stats.operations == {
        "test.test": {
            "attacksDetected": {"blocked": 1, "total": 1},
            "kind": "nosql_op",
            "total": 4,
        },
        "test.test2": {
            "kind": "sql_op",
            "total": 5,
            "attacksDetected": {"blocked": 5, "total": 200},
        },
    }


def test_empty(stats):
    assert stats.empty() == True

    stats.total_hits = 1
    assert stats.empty() == False

    stats.total_hits = 0
    stats.attacks_detected = 1
    assert stats.empty() == False

    stats.attacks_detected = 0
    stats.operations = {"test_op": {"total": 1}}
    assert stats.empty() == False


def test_multiple_imports(stats):
    record1 = {
        "requests": {
            "total": 10,
            "attacksDetected": {
                "total": 5,
                "blocked": 3,
            },
        },
        "operations": {
            "test_op1": {
                "total": 1,
                "kind": "fs_op",
                "attacksDetected": {"total": 0, "blocked": 0},
            }
        },
    }
    record2 = {
        "requests": {
            "total": 20,
            "attacksDetected": {
                "total": 10,
                "blocked": 7,
            },
        },
        "operations": {
            "test_op2": {
                "total": 1,
                "kind": "fs_op",
                "attacksDetected": {"total": 0, "blocked": 0},
            }
        },
    }
    stats.import_from_record(record1)
    stats.import_from_record(record2)
    assert stats.total_hits == 30
    assert stats.attacks_detected == 15
    assert stats.attacks_blocked == 10
    assert stats.operations == {
        "test_op1": {
            "attacksDetected": {"blocked": 0, "total": 0},
            "kind": "fs_op",
            "total": 1,
        },
        "test_op2": {
            "attacksDetected": {"blocked": 0, "total": 0},
            "kind": "fs_op",
            "total": 1,
        },
    }


def test_import_empty_record(stats):
    record = {"requests": {}}
    stats.import_from_record(record)
    assert stats.total_hits == 0
    assert stats.attacks_detected == 0
    assert stats.attacks_blocked == 0
    assert stats.operations == {}


def test_import_partial_record(stats):
    record = {"requests": {"total": 10}}
    stats.import_from_record(record)
    assert stats.total_hits == 10
    assert stats.attacks_detected == 0
    assert stats.attacks_blocked == 0
    assert stats.operations == {}


def test_increment_and_detect(stats):
    stats.increment_total_hits()
    stats.on_detected_attack(blocked=True, operation="test_op")
    assert stats.total_hits == 1
    assert stats.attacks_detected == 1
    assert stats.attacks_blocked == 1


def test_multiple_increments_and_detects(stats):
    stats.operations.register_call("test_op", "sql_op")
    for _ in range(10):
        stats.increment_total_hits()
    for _ in range(5):
        stats.on_detected_attack(blocked=True, operation="test_op")
    for _ in range(5):
        stats.on_detected_attack(blocked=False, operation="test_op")
    assert stats.total_hits == 10
    assert stats.attacks_detected == 10
    assert stats.attacks_blocked == 5
    assert stats.operations.get("test_op") == {
        "attacksDetected": {"blocked": 5, "total": 10},
        "kind": "sql_op",
        "total": 1,
    }



"""
patching module openai
- patches function create(...) on Responses class, to inspect response
- patches function create(...) on Completions class, to inspect response
"""

from aikido_zen.helpers.on_ai_call import on_ai_call
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import on_import, patch_function, after


@after
def _create_responses(func, instance, args, kwargs, return_value):
    op = f"openai.resources.responses.responses.Responses.{func.__name__}"
    register_call(op, "ai_op")

    on_ai_call(
        provider="openai",
        model=return_value.model,
        input_tokens=return_value.usage.input_tokens,
        output_tokens=return_value.usage.output_tokens,
    )


@after
def _create_completions(func, instance, args, kwargs, return_value):
    op = f"openai.resources.chat.completions.completions.Completions.{func.__name__}"
    register_call(op, "ai_op")

    on_ai_call(
        provider="openai",
        model=return_value.model,
        input_tokens=return_value.usage.prompt_tokens,
        output_tokens=return_value.usage.completion_tokens,
    )


@on_import("openai.resources.responses.responses", "openai", "1.0")
def patch_responses(m):
    patch_function(m, "Responses.create", _create_responses)
    patch_function(m, "Responses.parse", _create_responses)


@on_import("openai.resources.chat.completions.completions", "openai", "1.0")
def patch_chat_completions(m):
    patch_function(m, "Completions.create", _create_completions)
    patch_function(m, "Completions.update", _create_completions)



"""Mainly exports normalize_asgi_headers"""


def normalize_asgi_headers(headers):
    """
    Normalizes headers provided by ASGI :
    Decodes them, uppercase and underscore keys
    """
    parsed_headers = {}
    for k, v in headers:
        # Normalizing key : decoding, removing dashes and uppercase
        key_without_dashes = k.decode("utf-8").replace("-", "_")
        key_normalized = key_without_dashes.upper()
        parsed_headers[key_normalized] = v.decode("utf-8")
    return parsed_headers



import pytest
import random
from aikido_zen.helpers.looks_like_a_secret import looks_like_a_secret

LOWERCASE = "abcdefghijklmnopqrstuvwxyz"
UPPERCASE = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
NUMBERS = "0123456789"
SPECIALS = "!#$%^&*|;:<>"
MINIMUM_LENGTH = 10


def secret_from_charset(length, charset):
    return "".join(random.choice(charset) for _ in range(length))


def test_empty_string():
    assert looks_like_a_secret("") == False


def test_short_strings():
    short_strings = [
        "c",
        "NR",
        "7t3",
        "4qEK",
        "KJr6s",
        "KXiW4a",
        "Fupm2Vi",
        "jiGmyGfg",
        "SJPLzVQ8t",
        "OmNf04j6mU",
    ]
    for s in short_strings:
        assert looks_like_a_secret(s) == False


def test_long_strings():
    assert looks_like_a_secret("rsVEExrR2sVDONyeWwND") == True
    assert looks_like_a_secret(":2fbg;:qf$BRBc<2AG8&") == True


def test_very_long_strings():
    assert (
        looks_like_a_secret(
            "efDJHhzvkytpXoMkFUgag6shWJktYZ5QUrUCTfecFELpdvaoAT3tekI4ZhpzbqLt"
        )
        == True
    )
    assert (
        looks_like_a_secret(
            "XqSwF6ySwMdTomIdmgFWcMVXWf5L0oVvO5sIjaCPI7EjiPvRZhZGWx3A6mLl1HXPOHdUeabsjhngW06JiLhAchFwgtUaAYXLolZn75WsJVKHxEM1mEXhlmZepLCGwRAM"
        )
        == True
    )


def test_contains_white_space():
    assert looks_like_a_secret("rsVEExrR2sVDONyeWwND ") == False


def test_less_than_2_charsets():
    assert looks_like_a_secret(secret_from_charset(10, LOWERCASE)) == False
    assert looks_like_a_secret(secret_from_charset(10, UPPERCASE)) == False
    assert looks_like_a_secret(secret_from_charset(10, NUMBERS)) == False
    assert looks_like_a_secret(secret_from_charset(10, SPECIALS)) == False


def test_common_url_terms():
    url_terms = [
        "development",
        "programming",
        "applications",
        "implementation",
        "environment",
        "technologies",
        "documentation",
        "demonstration",
        "configuration",
        "administrator",
        "visualization",
        "international",
        "collaboration",
        "opportunities",
        "functionality",
        "customization",
        "specifications",
        "optimization",
        "contributions",
        "accessibility",
        "subscription",
        "subscriptions",
        "infrastructure",
        "architecture",
        "authentication",
        "sustainability",
        "notifications",
        "announcements",
        "recommendations",
        "communication",
        "compatibility",
        "enhancement",
        "integration",
        "performance",
        "improvements",
        "introduction",
        "capabilities",
        "communities",
        "credentials",
        "integration",
        "permissions",
        "validation",
        "serialization",
        "deserialization",
        "rate-limiting",
        "throttling",
        "load-balancer",
        "microservices",
        "endpoints",
        "data-transfer",
        "encryption",
        "authorization",
        "bearer-token",
        "multipart",
        "urlencoded",
        "api-docs",
        "postman",
        "json-schema",
        "serialization",
        "deserialization",
        "rate-limiting",
        "throttling",
        "load-balancer",
        "api-gateway",
        "microservices",
        "endpoints",
        "data-transfer",
        "encryption",
        "signature",
        "poppins-bold-webfont.woff2",
        "karla-bold-webfont.woff2",
        "startEmailBasedLogin",
        "jenkinsFile",
        "ConnectionStrings.config",
        "coach",
        "login",
        "payment_methods",
        "activity_logs",
        "feedback_responses",
        "balance_transactions",
        "customer_sessions",
        "payment_intents",
        "billing_portal",
        "subscription_items",
        "namedLayouts",
        "PlatformAction",
        "quickActions",
        "queryLocator",
        "relevantItems",
        "parameterizedSearch",
    ]
    for term in url_terms:
        assert looks_like_a_secret(term) == False


def test_known_word_separators():
    assert looks_like_a_secret("this-is-a-secret-1") == False


def test_number_is_not_a_secret():
    assert looks_like_a_secret("1234567890") == False
    assert looks_like_a_secret("1234567890" * 2) == False


def test_known_secrets():
    secrets = [
        "yqHYTS<agpi^aa1",
        "hIofuWBifkJI5iVsSNKKKDpBfmMqJJwuXMxau6AS8WZaHVLDAMeJXo3BwsFyrIIm",
        "AG7DrGi3pDDIUU1PrEsj",
        "CnJ4DunhYfv2db6T1FRfciRBHtlNKOYrjoz",
        "Gic*EfMq:^MQ|ZcmX:yW1",
        "AG7DrGi3pDDIUU1PrEsj",
    ]
    for secret in secrets:
        assert looks_like_a_secret(secret) == True



from aikido_zen.thread.thread_cache import get_cache


def on_ai_call(provider, model, input_tokens, output_tokens):
    cache = get_cache()
    if cache:
        cache.ai_stats.on_ai_call(provider, model, input_tokens, output_tokens)



"""
init.py file for the module to detect NoSQL Injections
"""

import json
from aikido_zen.helpers.is_mapping import is_mapping
from aikido_zen.helpers.build_path_to_payload import build_path_to_payload
from aikido_zen.helpers.try_decode_as_jwt import try_decode_as_jwt
from aikido_zen.context import UINPUT_SOURCES


def match_filter_part_in_user(user_input, filter_part, path_to_payload=None):
    """
    This tries to match a filter part to a part in user input
    """
    if not path_to_payload:
        path_to_payload = []
    if isinstance(user_input, str):
        jwt = try_decode_as_jwt(user_input)
        if jwt[0]:
            return match_filter_part_in_user(
                jwt[1], filter_part, path_to_payload + [{"type": "jwt"}]
            )

    if is_mapping(user_input):
        filtered_input = remove_keys_that_dont_start_with_dollar_sign(user_input)
        if filtered_input == filter_part:
            return {
                "match": True,
                "pathToPayload": build_path_to_payload(path_to_payload),
            }
        for key in user_input:
            match = match_filter_part_in_user(
                user_input[key],
                filter_part,
                path_to_payload + [{"type": "object", "key": key}],
            )
            if match.get("match"):
                return match
    if isinstance(user_input, list):
        for index, value in enumerate(user_input):
            match = match_filter_part_in_user(
                value,
                filter_part,
                path_to_payload + [{"type": "array", "index": index}],
            )
            if match.get("match"):
                return match

    return {"match": False}


def remove_keys_that_dont_start_with_dollar_sign(filter):
    """
    This removes key that don't start with $, since they are not dangerous
    """
    return {key: value for key, value in filter.items() if key.startswith("$")}


def find_filter_part_with_operators(user_input, part_of_filter):
    """
    This looks for parts in the filter that have NSQL operators (e.g. $)
    """
    if is_mapping(part_of_filter):
        obj = remove_keys_that_dont_start_with_dollar_sign(part_of_filter)
        if len(obj) > 0:
            result = match_filter_part_in_user(user_input, obj)

            if result.get("match"):
                return {
                    "found": True,
                    "pathToPayload": result.get("pathToPayload"),
                    "payload": obj,
                }
        for key in part_of_filter:
            result = find_filter_part_with_operators(user_input, part_of_filter[key])

            if result.get("found"):
                return {
                    "found": True,
                    "pathToPayload": result.get("pathToPayload"),
                    "payload": result.get("payload"),
                }

    if isinstance(part_of_filter, list):
        for val in part_of_filter:
            result = find_filter_part_with_operators(user_input, val)
            if result.get("found"):
                return {
                    "found": True,
                    "pathToPayload": result.get("pathToPayload"),
                    "payload": result.get("payload"),
                }

    return {"found": False}


def detect_nosql_injection(request, _filter):
    """
    Give a context object and a nosql filter and this function
    checks if there is a NoSQL injection
    """
    if not is_mapping(_filter) and not isinstance(_filter, list):
        return {}

    for source in UINPUT_SOURCES:
        if hasattr(request, source):
            result = find_filter_part_with_operators(getattr(request, source), _filter)

            if result.get("found"):
                return {
                    "injection": True,
                    "source": source,
                    "pathToPayload": result.get("pathToPayload"),
                    "payload": result.get("payload"),
                }

    return {}



"""
This will check the context of the request for shell injection
"""

from aikido_zen.helpers.extract_strings_from_context import extract_strings_from_context
from .detect_shell_injection import detect_shell_injection


def check_context_for_shell_injection(command, operation, context):
    """
    This will check the context of the request for Shell injections
    """
    if not isinstance(command, str):
        # Command must be string to run algorithm
        return {}
    for user_input, path, source in extract_strings_from_context(context):
        if detect_shell_injection(command, user_input):
            return {
                "operation": operation,
                "kind": "shell_injection",
                "source": source,
                "pathToPayload": path,
                "metadata": {"command": command},
                "payload": user_input,
            }
    return {}



"""Exports the contains_shell_syntax function"""

import regex as re

# Constants
dangerous_chars = [
    "#",
    "!",
    '"',
    "$",
    "&",
    "'",
    "(",
    ")",
    "*",
    ";",
    "<",
    "=",
    ">",
    "?",
    "[",
    "\\",
    "]",
    "^",
    "`",
    "{",
    "|",
    "}",
    " ",
    "\n",
    "\t",
    "~",
]

commands = [
    "sleep",
    "shutdown",
    "reboot",
    "poweroff",
    "halt",
    "ifconfig",
    "chmod",
    "chown",
    "ping",
    "ssh",
    "scp",
    "curl",
    "wget",
    "telnet",
    "kill",
    "killall",
    "rm",
    "mv",
    "cp",
    "touch",
    "echo",
    "cat",
    "head",
    "tail",
    "grep",
    "find",
    "awk",
    "sed",
    "sort",
    "uniq",
    "wc",
    "ls",
    "env",
    "ps",
    "who",
    "whoami",
    "id",
    "w",
    "df",
    "du",
    "pwd",
    "uname",
    "hostname",
    "netstat",
    "passwd",
    "arch",
    "printenv",
    "logname",
    "pstree",
    "hostnamectl",
    "set",
    "lsattr",
    "killall5",
    "dmesg",
    "history",
    "free",
    "uptime",
    "finger",
    "top",
    "shopt",
    ":",  # Colon is a null command
]

path_prefixes = [
    "/bin/",
    "/sbin/",
    "/usr/bin/",
    "/usr/sbin/",
    "/usr/local/bin/",
    "/usr/local/sbin/",
]

separators = [" ", "\t", "\n", ";", "&", "|", "(", ")", "<", ">"]


# Function to sort commands by length (longer commands first)
def by_length(a, b):
    """Sort by length desc"""
    return len(b) - len(a)


# Create the regex pattern
commands_regex = re.compile(
    r"([/.]*("
    + "|".join(map(re.escape, path_prefixes))
    + r")?("
    + "|".join(sorted(commands, key=len, reverse=True))
    + r"))",
    re.I | re.M,
)


def contains_shell_syntax(command, user_input):
    """
    Checks if the user input contains shell syntax given the command
    """
    if user_input.isspace():
        # The entire user input is just whitespace, ignore
        return False

    if any(c in user_input for c in dangerous_chars):
        return True

    # The command is the same as the user input
    # Rare case, but it's possible
    # e.g. command is `shutdown` and user input is `shutdown`
    # (`shutdown -h now` will be caught by the dangerous chars as it contains a space)
    if command == user_input:
        match = commands_regex.match(command)
        return match is not None and match.start() == 0 and match.end() == len(command)

    # Check if the command contains a commonly used command
    for match in commands_regex.finditer(command):
        # We found a command like `rm` or `/sbin/shutdown` in the command
        # Check if the command is the same as the user input
        # If it's not the same, continue searching
        if user_input != match[0]:
            continue

        # Otherwise, we'll check if the command is surrounded by separators
        # These separators are used to separate commands and arguments
        # e.g. `rm<space>-rf`
        # e.g. `ls<newline>whoami`
        # e.g. `echo<tab>hello`
        char_before = command[match.start() - 1] if match.start() > 0 else None
        char_after = command[match.end()] if match.end() < len(command) else None

        # Check surrounding characters
        if char_before in separators and char_after in separators:
            #  e.g. `<separator>rm<separator>`
            return True
        if char_before in separators and char_after is None:
            #  e.g. `<separator>rm`
            return True
        if char_before is None and char_after in separators:
            # e.g. `rm<separator>`
            return True
    return False



import pytest
from aikido_zen.helpers.is_mapping import is_mapping
from collections import defaultdict, OrderedDict, ChainMap


def test_is_mapping_true():
    assert is_mapping({}) == True
    assert is_mapping({"foo": "bar"}) == True


def test_is_mapping_false():
    class Foo:
        def __init__(self):
            self.abc = {}

    foo_instance = Foo()

    assert is_mapping("/foo/") == False
    assert is_mapping(lambda: None) == False
    assert is_mapping(1) == False
    assert is_mapping(["foo", "bar"]) == False
    assert is_mapping([]) == False
    assert is_mapping(foo_instance) == False
    assert is_mapping(None) == False


def test_is_mapping_modified_prototype():
    class CustomConstructor:
        pass

    CustomConstructor.prototype = list

    instance = CustomConstructor()

    assert is_mapping(instance) == False


def test_is_mapping_mapping_types():
    fruit_count = defaultdict(int)
    fruit_count["apple"] += 1
    fruit_count["banana"] += 2
    assert is_mapping(fruit_count) == True

    ordered_dict = OrderedDict()
    ordered_dict["first"] = 1
    ordered_dict["second"] = 2
    ordered_dict["third"] = 3
    assert is_mapping(ordered_dict) == True

    dict1 = {"a": 1, "b": 2}
    dict2 = {"b": 3, "c": 4}
    chain_map = ChainMap(dict1, dict2)
    assert is_mapping(chain_map) == True



import pytest
from aikido_zen.background_process.api import ReportingApi

# Test ReportingApi Class :
from requests.models import Response


@pytest.fixture
def reporting_api():
    return ReportingApi()


def test_to_api_response_rate_limited(reporting_api):
    res = Response()
    res.status_code = 429
    assert reporting_api.to_api_response(res) == {
        "success": False,
        "error": "rate_limited",
    }


def test_to_api_response_invalid_token(reporting_api):
    res = Response()
    res.status_code = 401
    assert reporting_api.to_api_response(res) == {
        "success": False,
        "error": "invalid_token",
    }


def test_to_api_response_unknown_error(reporting_api):
    res = Response()
    res.status_code = 500  # Simulating an unknown error status code
    assert reporting_api.to_api_response(res) == {
        "success": False,
        "error": "unknown_error",
    }


def test_to_api_response_valid_json(reporting_api):
    res = Response()
    res.status_code = 200
    res._content = b'{"key": "value"}'  # Simulating valid JSON response
    assert reporting_api.to_api_response(res) == {"key": "value"}



"""Exports run_init_stage function"""

import json
from aikido_zen.context import Context
from aikido_zen.helpers.logging import logger
from ..functions.request_handler import request_handler


def run_init_stage(request):
    """Parse request and body, run "init" stage with request_handler"""
    body = None
    try:
        # try-catch loading of form parameters, this is to fix issue with DATA_UPLOAD_MAX_NUMBER_FIELDS :
        try:
            body = request.POST.dict()
            if len(body) == 0:
                body = None  # Reset
        except Exception:
            pass

        # Check for JSON or XML :
        if body is None and request.content_type == "application/json":
            try:
                body = json.loads(request.body)
            except Exception:
                pass
        if body is None or len(body) == 0:
            # E.g. XML Data
            body = request.body
        if body is None or len(body) == 0:
            # During a GET request, django leaves the body as an empty byte string (e.g. `b''`).
            # When an attack is detected, this body needs to be serialized which would fail.
            # So a byte string gets converted into a string to stop that from happening.
            body = ""  # Set body to an empty string.
    except Exception as e:
        logger.debug("Exception occurred in run_init_stage function (Django) : %s", e)

    # In a separate try-catch we set the context :
    try:
        context = None
        if (
            hasattr(request, "scope") and request.scope is not None
        ):  # This request is an ASGI request
            context = Context(req=request.scope, body=body, source="django_async")
        elif hasattr(request, "META") and request.META is not None:  # WSGI request
            context = Context(req=request.META, body=body, source="django")
        else:
            return
        context.set_as_current_context()

        # Init stage needs to be run with context already set :
        request_handler(stage="init")
    except Exception as e:
        logger.debug("Exception occurred in run_init_stage function (Django): %s", e)



import os
import pytest
from unittest.mock import patch
from .get_api_url import get_api_url

DEFAULT_API_URL = "https://guard.aikido.dev/"


def test_get_api_url_no_env_var():
    """Test when AIKIDO_ENDPOINT is not set."""
    with patch.dict(os.environ, {}, clear=True):
        assert get_api_url() == DEFAULT_API_URL


def test_get_api_url_valid_url():
    """Test when AIKIDO_ENDPOINT is set to a valid URL."""
    valid_url = "https://example.com/api"
    with patch.dict(os.environ, {"AIKIDO_ENDPOINT": valid_url}):
        assert get_api_url() == valid_url + "/"


def test_get_api_url_invalid_url():
    """Test when AIKIDO_ENDPOINT is set to an invalid URL."""
    invalid_url = "invalid_url"
    with patch.dict(os.environ, {"AIKIDO_ENDPOINT": invalid_url}):
        assert get_api_url() == DEFAULT_API_URL


def test_get_api_url_valid_url_no_trailing_slash():
    """Test when AIKIDO_ENDPOINT is set to a valid URL without a trailing slash."""
    valid_url_no_slash = "https://example.com/api"
    with patch.dict(os.environ, {"AIKIDO_ENDPOINT": valid_url_no_slash}):
        assert get_api_url() == valid_url_no_slash + "/"


# You may need to mock the try_parse_url function if it has side effects or is complex



import pytest
from unittest.mock import patch
import aikido_zen.sinks.builtins
from pathlib import Path, PurePath

kind = "path_traversal"
op = "builtins.open"


def test_open():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        with pytest.raises(FileNotFoundError):
            open("test_file")
        args = ("test_file",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)
        with pytest.raises(FileNotFoundError):
            open("")
        args = ("",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            open("ltlwtjl_tlnekt.py")
        args = ("ltlwtjl_tlnekt.py",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            open(b"afljqlqfefjq.py")
        args = (b"afljqlqfefjq.py",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        path = Path("./test", "/test.py")
        with pytest.raises(FileNotFoundError):
            open(path)
        args = (path,)
        # Need to use assert_any_call, since python 3.12 it uses os.path.join
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        path2 = PurePath("./test", "/test.py")
        with pytest.raises(FileNotFoundError):
            open(path2)
        args = (path2,)
        # Need to use assert_any_call, since python 3.12 it uses os.path.join
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)


def test_open_with_builtins_import():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import builtins

        with pytest.raises(FileNotFoundError):
            builtins.open("test_file")
        args = ("test_file",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)
        with pytest.raises(FileNotFoundError):
            builtins.open("")
        args = ("",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            builtins.open("ltlwtjl_tlnekt.py")
        args = ("ltlwtjl_tlnekt.py",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(FileNotFoundError):
            builtins.open(b"shleklelkwge.py")
        args = (b"shleklelkwge.py",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)


def test_open_invalid_input():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        with pytest.raises(TypeError):
            open()
        mock_run_vulnerability_scan.assert_not_called()

        with pytest.raises(TypeError):
            open(123456789123456789)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        with pytest.raises(TypeError):
            open(None)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        with pytest.raises(TypeError):
            open(["list", "of", "commands"])
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test dictionary command
        with pytest.raises(TypeError):
            open({"key": "value"})
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test float command
        with pytest.raises(TypeError):
            open(3.14)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test boolean command
        open(True)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test tuple command
        with pytest.raises(TypeError):
            open(("tuple", "command"))
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called



import pytest
from aikido_zen.helpers.build_path_to_payload import build_path_to_payload


def test_build_path_to_payload_empty():
    assert build_path_to_payload([]) == "."


def test_build_path_to_payload_single_object():
    path = [{"type": "object", "key": "name"}]
    assert build_path_to_payload(path) == ".name"


def test_build_path_to_payload_single_array():
    path = [{"type": "array", "index": 0}]
    assert build_path_to_payload(path) == ".[0]"


def test_build_path_to_payload_single_jwt():
    path = [{"type": "jwt"}]
    assert build_path_to_payload(path) == "<jwt>"


def test_build_path_to_payload_mixed_types():
    path = [
        {"type": "object", "key": "user"},
        {"type": "array", "index": 2},
        {"type": "jwt"},
        {"type": "object", "key": "details"},
        {"type": "array", "index": 1},
    ]
    assert build_path_to_payload(path) == ".user.[2]<jwt>.details.[1]"


def test_build_path_to_payload_multiple_objects():
    path = [
        {"type": "object", "key": "user"},
        {"type": "object", "key": "details"},
        {"type": "object", "key": "address"},
    ]
    assert build_path_to_payload(path) == ".user.details.address"



"""Exports `build_url_from_wsgi`"""


def build_url_from_wsgi(request):
    """Builds a URL from the different parts in a WSGI request"""
    scheme = request["wsgi.url_scheme"]
    uri = request["PATH_INFO"]

    host = ""
    if "HTTP_HOST" not in request or request["HTTP_HOST"] == "":
        host = request["SERVER_NAME"]
        port = request["SERVER_PORT"]

        host = f"{host}:{port}"
    else:
        host = request["HTTP_HOST"]

    return f"{scheme}://{host}{uri}"



"""Exports request_handler function"""

import aikido_zen.context as ctx
from aikido_zen.api_discovery.update_route_info import update_route_info_from_context
from aikido_zen.helpers.is_useful_route import is_useful_route
from aikido_zen.helpers.logging import logger
from aikido_zen.thread.thread_cache import get_cache
from .ip_allowed_to_access_route import ip_allowed_to_access_route
from ...helpers.is_ip_allowed_by_allowlist import is_ip_allowed_by_allowlist


def request_handler(stage, status_code=0):
    """This will check for rate limiting, Allowed IP's, useful routes, etc."""
    try:
        if stage == "init":
            cache = get_cache()
            if ctx.get_current_context() and cache:
                cache.stats.increment_total_hits()
        if stage == "pre_response":
            return pre_response()
        if stage == "post_response":
            return post_response(status_code)
    except Exception as e:
        logger.debug("Exception occurred in request_handler : %s", e)
    return None


def pre_response():
    """
    This is executed at the end of the middleware chain before a response is present
    - Checks if the IP is allowed to access the route (route-specific)
    - Checks if the IP is in an allowlist (if that exists)
    - Checks if the IP is in a blocklist (if that exists)
    - Checks if the user agent is blocked (if the regex exists)
    """
    context = ctx.get_current_context()
    cache = get_cache()
    if not context or not cache:
        logger.debug("Request was not complete, not running any pre_response code")
        return

    # Per endpoint IP Allowlist
    matched_endpoints = cache.config.get_endpoints(context.get_route_metadata())
    if not ip_allowed_to_access_route(
        context.remote_address, context.get_route_metadata(), matched_endpoints
    ):
        message = "Your IP address is not allowed to access this resource."
        if context.remote_address:
            message += f" (Your IP: {context.remote_address})"
        return message, 403

    # Global IP Allowlist (e.g. for geofencing)
    if not is_ip_allowed_by_allowlist(cache.config, context.remote_address):
        message = "Your IP address is not allowed."
        message += " (Your IP: " + context.remote_address + ")"
        return message, 403

    # Global IP Blocklist (e.g. blocking known threat actors)
    reason = cache.config.is_blocked_ip(context.remote_address)
    if reason:
        message = "Your IP address is blocked due to " + reason
        message += " (Your IP: " + context.remote_address + ")"
        return message, 403

    # User agent blocking (e.g. blocking AI scrapers)
    if cache.config.is_user_agent_blocked(context.get_user_agent()):
        msg = "You are not allowed to access this resource because you have been identified as a bot."
        return msg, 403


def post_response(status_code):
    """Checks if the current route is useful"""
    context = ctx.get_current_context()
    if not context:
        return
    route_metadata = context.get_route_metadata()

    is_curr_route_useful = is_useful_route(
        status_code,
        context.route,
        context.method,
    )
    if not is_curr_route_useful:
        return

    cache = get_cache()
    if cache:
        cache.routes.increment_route(route_metadata)

        # api spec generation
        route = cache.routes.get(route_metadata)
        update_route_info_from_context(context, route)



import pytest
from .extract_route_params import extract_route_params


def test_with_urlencoded_urls():
    url1 = "http://localhost:8080/app/shell/ls%20-la"
    assert extract_route_params(url1) == ["ls -la", "app/shell/ls%20-la"]

    url2 = "http://localhost:8080/app/shell/ls -la"
    assert extract_route_params(url2) == ["ls -la", "app/shell/ls -la"]


def test_uses_keys():
    url = "http://localhost:8080/app/shell/me@woutfeys.be/017shell/127.0.0.1/"
    assert extract_route_params(url) == [
        "me@woutfeys.be",
        "127.0.0.1",
        "app/shell/me@woutfeys.be/017shell/127.0.0.1/",
    ]


def test_normal_urls():
    assert extract_route_params("http://localhost:8080/a/b/abc2393027def/def") == []


def test_with_empty_route():
    url1 = "http://localhost:8080"
    assert extract_route_params(url1) == []

    url2 = "http://localhost:8080"
    assert extract_route_params(url2) == []


def test_special_characters():
    url1 = "http://localhost:8080/app/shell/!@#$%^&*()"  # Everything past hashtag is not url anymore
    assert extract_route_params(url1) == ["!@", "app/shell/!@"]

    url2 = "http://localhost:8080/app/shell/space test"
    assert extract_route_params(url2) == ["space test", "app/shell/space test"]

    url3 = "http://localhost:8080/app/shell/hello%20world"
    assert extract_route_params(url3) == ["hello world", "app/shell/hello%20world"]


def test_numeric_segments():
    # Alphanum is ignored:
    url1 = "http://localhost:8080/app/shell/12345"
    assert extract_route_params(url1) == []

    url2 = "http://localhost:8080/app/shell/67890/abc"
    assert extract_route_params(url2) == []


def test_mixed_segments():
    url1 = "http://localhost:8080/app/shell/abc123/!@#"
    assert extract_route_params(url1) == ["!@", "app/shell/abc123/!@"]

    url2 = "http://localhost:8080/app/shell/abc/123/!@#"
    assert extract_route_params(url2) == ["!@", "app/shell/abc/123/!@"]


def test_encoded_and_unencoded():
    url1 = "http://localhost:8080/app/shell/%E2%9C%93"
    assert extract_route_params(url1) == ["✓", "app/shell/%E2%9C%93"]

    url2 = "http://localhost:8080/app/shell/%E2%9C%93/normal"
    assert extract_route_params(url2) == ["✓", "app/shell/%E2%9C%93/normal"]


def test_no_params():
    url1 = "http://localhost:8080/app/shell/"
    assert extract_route_params(url1) == []

    url2 = "http://localhost:8080/app/"
    assert extract_route_params(url2) == []


def test_edge_cases():
    url1 = "http://localhost:8080/app/shell/.."
    assert extract_route_params(url1) == ["..", "app/shell/.."]

    url2 = "http://localhost:8080/app/shell/./"
    assert extract_route_params(url2) == ["app/shell/./"]


def test_long_urls():
    url1 = "http://localhost:8080/app./shell/" + "a" * 1000
    assert extract_route_params(url1) == ["app.", "app./shell/" + "a" * 1000]

    url2 = "http://localhost:8080/app./shell/" + "b" * 1000 + "/c" * 1000
    assert extract_route_params(url2) == [
        "app.",
        "app./shell/" + "b" * 1000 + "/c" * 1000,
    ]


def test_query_parameters():
    # Test query parameters are ignored:
    url1 = "http://localhost:8080/app/./shell/?param=value"
    assert extract_route_params(url1) == ["app/./shell/"]

    url2 = "http://localhost:8080/app/./shell/?key1=value1&key2=value2"
    assert extract_route_params(url2) == ["app/./shell/"]


def test_fragment_identifiers():
    # Fragments should be ignored:
    url1 = "http://localhost:8080/app/./shell/#section1"
    assert extract_route_params(url1) == ["app/./shell/"]

    url2 = "http://localhost:8080/app/shell/#/path/to/resource"
    assert extract_route_params(url2) == []



"""
Exports `run_vulnerability_scan` function
"""

from aikido_zen.helpers.serialize_to_json import serialize_to_json
from aikido_zen.context import get_current_context
from aikido_zen.errors import (
    AikidoException,
    AikidoSQLInjection,
    AikidoNoSQLInjection,
    AikidoShellInjection,
    AikidoPathTraversal,
    AikidoSSRF,
)
import aikido_zen.background_process.comms as comm
from aikido_zen.helpers.logging import logger
from aikido_zen.helpers.get_clean_stacktrace import get_clean_stacktrace
from aikido_zen.helpers.blocking_enabled import is_blocking_enabled
from aikido_zen.helpers.protection_forced_off import protection_forced_off
from aikido_zen.thread.thread_cache import get_cache
from .sql_injection.context_contains_sql_injection import context_contains_sql_injection
from .nosql_injection.check_context import check_context_for_nosql_injection
from .ssrf.inspect_getaddrinfo_result import inspect_getaddrinfo_result
from .shell_injection.check_context_for_shell_injection import (
    check_context_for_shell_injection,
)
from .path_traversal.check_context_for_path_traversal import (
    check_context_for_path_traversal,
)


def run_vulnerability_scan(kind, op, args):
    """
    Generally checks context for the provided "kind" of injection,
    raises error if blocking is enabled, communicates it with connection_manager
    """
    context = get_current_context()
    comms = comm.get_comms()
    thread_cache = get_cache()
    if not context and kind != "ssrf":
        # Make a special exception for SSRF, which checks itself if context is set.
        # This is because some scans/tests for SSRF do not require a context to be set.
        return

    if not thread_cache and kind != "ssrf":
        # Make a special exception for SSRF, which checks itself if thread cache is set.
        # This is because some scans/tests for SSRF do not require a thread cache to be set.
        return
    if thread_cache and context:
        if protection_forced_off(
            context.get_route_metadata(), thread_cache.get_endpoints()
        ):
            #  The client turned protection off for this route, not scanning
            return
        if thread_cache.is_bypassed_ip(context.remote_address):
            #  This IP is on the bypass list, not scanning
            return

    error_type = AikidoException  # Default error
    error_args = tuple()
    injection_results = {}
    try:
        if kind == "sql_injection":
            injection_results = context_contains_sql_injection(
                sql=args[0], dialect=args[1], operation=op, context=context
            )
            error_type = AikidoSQLInjection
            error_args = (args[1],)  # Pass along the dialect
        elif kind == "nosql_injection":
            injection_results = check_context_for_nosql_injection(
                context=context, op=op, _filter=args[0]
            )
            error_type = AikidoNoSQLInjection
        elif kind == "shell_injection":
            injection_results = check_context_for_shell_injection(
                command=args[0], operation=op, context=context
            )
            error_type = AikidoShellInjection
        elif kind == "path_traversal":
            injection_results = check_context_for_path_traversal(
                filename=args[0], operation=op, context=context
            )
            error_type = AikidoPathTraversal
        elif kind == "ssrf":
            dns_results, hostname, port = args
            injection_results = inspect_getaddrinfo_result(dns_results, hostname, port)
            error_type = AikidoSSRF

            if thread_cache and port > 0:
                thread_cache.hostnames.add(hostname, port)
        else:
            logger.error(
                "Vulnerability type %s currently has no scans implemented", kind
            )
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.debug("Exception occurred in run_vulnerability_scan : %s", e)

    if injection_results:
        logger.debug("Injection results : %s", serialize_to_json(injection_results))

        blocked = is_blocking_enabled()
        operation = injection_results["operation"]
        thread_cache.stats.on_detected_attack(blocked, operation)

        stack = get_clean_stacktrace()

        if comms:
            comms.send_data_to_bg_process(
                "ATTACK", (injection_results, context, blocked, stack)
            )
        if blocked:
            raise error_type(*error_args)



import pytest
from unittest.mock import MagicMock, patch
from .is_redirect_to_private_ip import is_redirect_to_private_ip
from urllib.parse import urlparse, urlunparse


# Helper function to create URL objects
def create_url(href):
    return urlparse(href)


def test_is_redirect_to_private_ip_success():
    context = MagicMock()
    context.outgoing_req_redirects = [
        {
            "source": create_url("http://example.com"),
            "destination": create_url("http://192.168.0.1/"),
        },
    ]
    context.parsed_userinput = {}
    context.body = {"field": ["http://example.com"]}
    with patch("aikido_zen.context.get_current_context", return_value=context):
        result = is_redirect_to_private_ip("192.168.0.1", context, 80)
        assert result == {
            "pathToPayload": ".field.[0]",
            "payload": "http://example.com",
            "source": "body",
        }


def test_is_redirect_to_private_ip_no_redirects():
    context = MagicMock()
    context.outgoing_req_redirects = []

    result = is_redirect_to_private_ip("192.168.0.1", context, 80)
    assert result is None


def test_is_redirect_to_private_ip_not_private_ip():
    context = MagicMock()
    context.outgoing_req_redirects = [
        {
            "source": create_url("http://example.com"),
            "destination": create_url("http://hackers.com"),
        },
    ]

    result = is_redirect_to_private_ip("example.com", context, 443)
    assert result is None


def test_is_redirect_to_private_ip_redirect_origin_not_found():
    context = MagicMock()
    context.outgoing_req_redirects = [
        {
            "source": create_url("http://example.com"),
            "destination": create_url("http://notfound.com"),
        },
    ]

    with MagicMock() as mock_get_redirect_origin:
        mock_get_redirect_origin.return_value = None

        with pytest.MonkeyPatch.context() as mp:
            mp.setattr(
                "aikido_zen.vulnerabilities.ssrf.get_redirect_origin",
                mock_get_redirect_origin,
            )

            result = is_redirect_to_private_ip("192.168.0.1", context, 80)
            assert result is None


def test_is_redirect_to_private_ip_hostname_not_found_in_context():
    context = MagicMock()
    context.outgoing_req_redirects = [
        {
            "source": create_url("http://example.com"),
            "destination": create_url("http://hackers.com"),
        },
    ]

    with MagicMock() as mock_get_redirect_origin, MagicMock() as mock_find_hostname_in_context:
        mock_get_redirect_origin.return_value = MagicMock(
            hostname="example.com", port=80
        )
        mock_find_hostname_in_context.return_value = False

        with pytest.MonkeyPatch.context() as mp:
            mp.setattr(
                "aikido_zen.vulnerabilities.ssrf.get_redirect_origin",
                mock_get_redirect_origin,
            )
            mp.setattr(
                "aikido_zen.vulnerabilities.ssrf.find_hostname_in_context",
                mock_find_hostname_in_context,
            )

            result = is_redirect_to_private_ip("192.168.0.1", context, 80)
            assert result is None



import pytest
from .get_argument import get_argument


def test_get_argument_with_only_kwargs():
    """Test when only kwargs are provided."""
    result = get_argument((), {"arg1": "value1"}, 0, "arg1")
    assert result == "value1", f"Expected 'value1', got {result}"


def test_get_argument_with_only_args():
    """Test when only args are provided."""
    result = get_argument(("value2",), {}, 0, "arg1")
    assert result == "value2", f"Expected 'value2', got {result}"


def test_get_argument_with_args_and_kwargs():
    """Test when both args and kwargs are provided, with priority to kwargs."""
    result = get_argument(("value2",), {"arg1": "value1"}, 0, "arg1")
    assert result == "value1", f"Expected 'value1', got {result}"


def test_get_argument_with_positional_index():
    """Test when args are provided and a specific position is requested."""
    result = get_argument(("value2", "value3"), {}, 1, "arg1")
    assert result == "value3", f"Expected 'value3', got {result}"


def test_get_argument_with_positional_index_out_of_bounds():
    """Test when the positional index is out of bounds."""
    result = get_argument(("value2",), {}, 1, "arg1")
    assert result is None, f"Expected None, got {result}"


def test_get_argument_with_none_in_kwargs():
    """Test when the argument in kwargs is None."""
    result = get_argument((), {"arg1": None}, 0, "arg1")
    assert result is None, f"Expected None, got {result}"


def test_get_argument_with_none_in_args():
    """Test when the argument in args is None."""
    result = get_argument((None,), {}, 0, "arg1")
    assert result is None, f"Expected None, got {result}"


def test_get_argument_with_empty_args_and_kwargs():
    """Test when both args and kwargs are empty."""
    result = get_argument((), {}, 0, "arg1")
    assert result is None, f"Expected None, got {result}"


def test_get_argument_with_multiple_kwargs():
    """Test when multiple kwargs are provided."""
    result = get_argument((), {"arg1": "value1", "arg2": "value2"}, 0, "arg1")
    assert result == "value1", f"Expected 'value1', got {result}"


def test_get_argument_with_positional_index_and_kwargs():
    """Test when both args and kwargs are provided, with positional index."""
    result = get_argument(("value2", "value3"), {"arg1": "value1"}, 0, "arg1")
    assert result == "value1", f"Expected 'value1', got {result}"



import multiprocessing
import time

from aikido_zen.helpers.logging import logger
from aikido_zen.thread import thread_cache

# Renew the cache from this background worker every 5 seconds
RENEW_CACHE_EVERY_X_SEC = 5


def aikido_process_worker_thread():
    """
    process worker -> When a web server like gUnicorn makes new processes, and those have multiple threads,
    Aikido process worker is linked to those new processes, so in essence it's 1 extra thread. This thread
    is responsible for syncing statistics, route data, configuration, ...
    """
    # Get the current process
    current_process = multiprocessing.current_process()

    while True:
        # Print information about the process
        logger.debug(
            f"Process ID: {current_process.pid}, Name: {current_process.name} - process_worker renewing thread cache."
        )

        # Renew the cache every 5 seconds
        thread_cache.renew()
        time.sleep(RENEW_CACHE_EVERY_X_SEC)



"""
Exports the HTTP API class
"""

import requests
from aikido_zen.background_process.api import ReportingApi
from aikido_zen.helpers.logging import logger


class ReportingApiHTTP(ReportingApi):
    """HTTP Reporting API"""

    def __init__(self, reporting_url):
        self.reporting_url = reporting_url

    def report(self, token, event, timeout_in_sec):
        try:
            res = requests.post(
                self.reporting_url + "api/runtime/events",
                json=event,
                timeout=timeout_in_sec,
                headers=get_headers(token),
            )
        except requests.exceptions.ConnectionError as e:
            logger.error(e)
            return {"success": False, "error": "timeout"}
        except Exception as e:
            logger.error(e)
            return {"success": False, "error": "unknown"}
        return self.to_api_response(res)

    def fetch_firewall_lists(self, token):
        """
        Fetches firewall lists from aikido's servers
        If successful the current API returns :
        - `allowedIPAddresses` : An array with iplist entries which are the only ones allowed
        - `blockedIPAddresses` : An array with iplist entries that are blocked
        - `blockedUserAgents` : A string with a simple regex to match user agents with
        """
        try:
            res = requests.get(
                self.reporting_url + "api/runtime/firewall/lists",
                timeout=20,
                headers={
                    # We need to set the Accept-Encoding header to "gzip" to receive the response in gzip format
                    "Accept-Encoding": "gzip",
                    "Authorization": str(token),
                },
            )
        except requests.exceptions.ConnectionError as e:
            logger.error(e)
            return {"success": False, "error": "timeout"}
        except Exception as e:
            logger.error(e)
            return {"success": False, "error": "unknown"}
        return self.to_api_response(res)


def get_headers(token):
    """Returns headers"""
    return {"Content-Type": "application/json", "Authorization": str(token)}



from aikido_zen.helpers.get_ip_from_request import trust_proxy
from aikido_zen.helpers.get_port_from_url import get_port_from_url
from aikido_zen.helpers.try_parse_url import try_parse_url


def is_request_to_itself(server_url, outbound_hostname, outbound_port):
    """
    We don't want to block outgoing requests to the same host as the server
    (often happens that we have a match on headers like `Host`, `Origin`, `Referer`, etc.)
    We have to check the port as well, because the hostname can be the same but with a different port
    """

    # When the app is not behind a reverse proxy, we can't trust the hostname inside `server_url`
    # The hostname in `server_url` is built from the request headers
    # The headers can be manipulated by the client if the app is directly exposed to the internet
    if not trust_proxy():
        return False

    base_url = try_parse_url(server_url)
    if base_url is None:
        return False

    if base_url.hostname != outbound_hostname:
        return False

    base_url_port = get_port_from_url(base_url, parsed=True)

    # If the port and hostname are the same, the server is making a request to itself
    if base_url_port == outbound_port:
        return True

    # Special case for HTTP/HTTPS ports
    # In production, the app will be served on port 80 and 443
    if base_url_port == 80 and outbound_port == 443:
        return True
    if base_url_port == 443 and outbound_port == 80:
        return True

    return False



"""
Exports map_dialect_to_rust_int
"""

DIALECTS = {
    "generic": 0,
    "clickhouse": 3,
    "mysql": 8,
    "postgres": 9,
    "sqlite": 12,
}


def map_dialect_to_rust_int(dialect):
    """
    This takes the string dialect as input and maps it to a rust integer
    Reference : [rust lib]/src/sql_injection/helpers/select_dialect_based_on_enum.rs
    """
    return DIALECTS[dialect]



"""
Holds the globally stored comms object
Exports the AikidoIPCCommunications class
"""

import multiprocessing.connection as con
from threading import Thread
from aikido_zen.helpers.logging import logger

# pylint: disable=invalid-name # This variable does change
comms = None


def get_comms():
    """
    Returns the globally stored IPC object, which you need
    to communicate to our background process.
    """
    return comms


def reset_comms():
    """This will reset communications"""
    # pylint: disable=global-statement # This needs to be global
    global comms
    if comms:
        logger.debug("Resetting communications. (comms = None)")
        comms = None


class AikidoIPCCommunications:
    """
    Facilitates Inter-Process communication
    """

    def __init__(self, address, key):
        # The key needs to be in byte form
        self.address = address
        self.key = key

        # Set as global ipc object :
        reset_comms()
        # pylint: disable=global-statement # This needs to be global
        global comms
        comms = self

    def send_data_to_bg_process(self, action, obj, receive=False):
        """Try-catched send_data_to_bg_process"""
        try:
            return self._send_data_to_bg_process(action, obj, receive)
        except Exception as e:
            logger.debug("Exception happened in send_data_to_bg_process : %s", e)
            return {"success": False, "error": "unknown"}

    def _send_data_to_bg_process(self, action, obj, receive=False):
        """
        This creates a new client for comms to the background process
        """
        if not self.key:
            # If no key is set, the background process will not start
            return {"success": False, "error": "invalid_key"}

        # We want to make sure that sending out this data affects the process as little as possible
        # So we run it inside a separate thread with a timeout of 100ms
        # If something goes wrong, it will also be encapsulated in the thread i.e. no crashes
        def target(address, key, receive, data, result_obj):
            try:
                # Create a connection, this can get stuck :
                conn = con.Client(address, authkey=None)

                # Send/Receive data :
                conn.send(data)
                if receive:
                    result_obj[1] = conn.recv()
                # Close the connection :
                conn.close()
                result_obj[0] = True  #  Connection ended gracefully
            except Exception as e:
                logger.debug("Exception occurred in thread : %s", e)

        # Create a shared result object between the thread and this process :
        result_obj = [False, None]  # Needs to be an array so we can make a ref.
        t = Thread(
            target=target,
            args=(self.address, self.key, receive, (action, obj), result_obj),
            daemon=True,  #  This allows us to join and set a timeout after which the thread closes
        )

        # Start and join the thread for 100ms, afterwards the thread is forced to close (daemon=True)
        t.start()
        t.join(timeout=0.1)
        if not result_obj[0]:
            logger.debug(
                " Failure in communication to background process, %s(%s)", action, obj
            )
            return {"success": False, "error": "timeout"}

        if receive:
            return {"success": True, "data": result_obj[1]}
        return {"success": True, "data": None}



"""Exports blocklist class and get_ip_address_type"""

import ipaddress


def get_ip_address_type(ip):
    """Returns type, ipv4 or ipv6"""
    try:
        if ipaddress.ip_address(ip).version == 4:
            return "ipv4"
        elif ipaddress.ip_address(ip).version == 6:
            return "ipv6"
    except ValueError:
        return None


class IPList:
    """A blocklist, where you can add subnets and addresses"""

    def __init__(self, ip_list=None):
        """Initializes the IPList, optionally with a list of IPs or CIDRs"""
        self.blocked_addresses = set()
        self.blocked_subnets = []

        if ip_list:
            for ip_or_cidr in ip_list:
                self.add(ip_or_cidr)

    def add_address(self, ip: str, ip_type: str):
        self.blocked_addresses.add((ip, ip_type))

    def add_subnet(self, plain_ip: str, ip_range: int, ip_type: str):
        self.blocked_subnets.append((plain_ip, ip_range, ip_type))

    def add(self, ip_or_cidr):
        """
        Checks whether ip_or_cidr is an IP address or is a subnet, and decides the correct IP type (IPv4 or IPv6)
        """
        if "/" not in ip_or_cidr:  # IP Address
            ip_type = get_ip_address_type(ip_or_cidr)
            if ip_type:
                self.add_address(ip_or_cidr, ip_type)
        else:  # Subnet
            plain_ip, range_str = ip_or_cidr.split("/")
            try:
                ip_range = int(range_str)
            except ValueError:
                return
            ip_type = get_ip_address_type(plain_ip)
            if not ip_type:
                return

            self.add_subnet(plain_ip, ip_range, ip_type)

    def matches(self, ip: str) -> bool:
        ip_type = get_ip_address_type(ip)
        if not ip_type:
            return False

        # Check if the IP address is in the blocked addresses
        if (ip, ip_type) in self.blocked_addresses:
            return True

        # Check if the IP address is in any of the blocked subnets
        ip_addr = ipaddress.ip_address(ip)
        for plain_ip, ip_range, _ in self.blocked_subnets:
            subnet = ipaddress.ip_network(f"{plain_ip}/{ip_range}", strict=False)
            if ip_addr in subnet:
                return True

        return False



import pytest
import socket
from unittest.mock import MagicMock, patch
from .get_manager_info import get_manager_info


# Sample connection_manager object for testing
class MockCloudConnectionManager:
    def __init__(self, block, packages, serverless):
        self.block = block
        self.packages = packages
        self.serverless = serverless


@pytest.fixture
def mock_connection_manager():
    return MockCloudConnectionManager(
        block=True,
        packages={
            "package1": {"version": "1.0.0", "supported": True},
            "package2": {"version": "2.0.0", "supported": False},
            "package3": {"version": "3.0.0", "supported": True},
        },
        serverless=True,
    )


# Fixture to patch socket.gethostname
@pytest.fixture(autouse=True)
def patch_gethostname(monkeypatch):
    monkeypatch.setattr(socket, "gethostname", lambda: "test-hostname")


@patch("platform.system", return_value="Linux")
@patch("platform.release", return_value="5.4.0")
@patch("aikido_zen.helpers.get_machine_ip.get_ip", return_value="192.168.1.1")
@patch("aikido_zen.config.PKG_VERSION", "PKG_VERSION")
def test_get_manager_info(
    mock_get_ip,
    mock_platform_release,
    mock_platform_system,
    mock_connection_manager,
    monkeypatch,
):
    connection_manager_info = get_manager_info(mock_connection_manager)
    print(connection_manager_info)
    assert connection_manager_info["dryMode"] is False
    assert connection_manager_info["hostname"] == "test-hostname"
    assert (
        connection_manager_info["version"] == "PKG_VERSION"
    )  # Replace with actual version if needed
    assert connection_manager_info["library"] == "firewall-python"
    assert connection_manager_info["ipAddress"] == "192.168.1.1"
    assert connection_manager_info["packages"] == {
        "package1": "1.0.0",
        "package3": "3.0.0",
    }
    assert connection_manager_info["serverless"] is True
    assert connection_manager_info["stack"] == [
        "package1",
        "package2",
        "package3",
        True,
    ]
    assert connection_manager_info["os"] == {"name": "Linux", "version": "5.4.0"}
    assert connection_manager_info["preventedPrototypePollution"] is False
    assert connection_manager_info["nodeEnv"] == ""


def test_get_manager_info_with_empty_packages():
    mock_connection_manager = MockCloudConnectionManager(
        block=False, packages={}, serverless=False
    )
    with patch("platform.system", return_value="Linux"), patch(
        "platform.release", return_value="5.4.0"
    ), patch("aikido_zen.helpers.get_machine_ip.get_ip", return_value="192.168.1.1"):
        connection_manager_info = get_manager_info(mock_connection_manager)

        assert connection_manager_info["packages"] == {}
        assert connection_manager_info["stack"] == []


def test_get_manager_info_with_unsupported_packages():
    mock_connection_manager = MockCloudConnectionManager(
        block=False,
        packages={
            "package1": {"version": "1.0.0", "supported": True},
            "package2": {"version": "2.0.0", "supported": False},
        },
        serverless=False,
    )

    with patch("platform.system", return_value="Linux"), patch(
        "platform.release", return_value="5.4.0"
    ), patch("aikido_zen.helpers.get_machine_ip.get_ip", return_value="192.168.1.1"):
        connection_manager_info = get_manager_info(mock_connection_manager)

        assert connection_manager_info["packages"] == {"package1": "1.0.0"}
        assert connection_manager_info["stack"] == ["package1", "package2"]


def test_get_manager_info_with_non_serverless():
    mock_connection_manager = MockCloudConnectionManager(
        block=False,
        packages={"package1": {"version": "1.0.0", "supported": True}},
        serverless=False,
    )

    with patch("platform.system", return_value="Linux"), patch(
        "platform.release", return_value="5.4.0"
    ), patch("aikido_zen.helpers.get_machine_ip.get_ip", return_value="192.168.1.1"):
        connection_manager_info = get_manager_info(mock_connection_manager)

        assert connection_manager_info["serverless"] is False
        assert connection_manager_info["stack"] == ["package1"]


def test_get_manager_info_with_blocked_connection_manager():
    mock_connection_manager = MockCloudConnectionManager(
        block=True,
        packages={"package1": {"version": "1.0.0", "supported": True}},
        serverless=False,
    )

    with patch("platform.system", return_value="Linux"), patch(
        "platform.release", return_value="5.4.0"
    ), patch("aikido_zen.helpers.get_machine_ip.get_ip", return_value="192.168.1.1"):
        connection_manager_info = get_manager_info(mock_connection_manager)

        assert connection_manager_info["dryMode"] is False


def test_get_manager_info_with_drymode_connection_manager():
    mock_connection_manager = MockCloudConnectionManager(
        block=False,
        packages={"package1": {"version": "1.0.0", "supported": True}},
        serverless=False,
    )

    with patch("platform.system", return_value="Linux"), patch(
        "platform.release", return_value="5.4.0"
    ), patch("aikido_zen.helpers.get_machine_ip.get_ip", return_value="192.168.1.1"):
        connection_manager_info = get_manager_info(mock_connection_manager)

        assert connection_manager_info["dryMode"] is True


def test_get_manager_info_with_different_os():
    mock_connection_manager = MockCloudConnectionManager(
        block=False,
        packages={"package1": {"version": "1.0.0", "supported": True}},
        serverless=False,
    )

    with patch("platform.system", return_value="Windows"), patch(
        "platform.release", return_value="10"
    ), patch("aikido_zen.helpers.get_machine_ip.get_ip", return_value="192.168.1.1"):
        connection_manager_info = get_manager_info(mock_connection_manager)

        assert connection_manager_info["os"] == {"name": "Windows", "version": "10"}


def test_get_manager_info_python_implementation():
    mock_connection_manager = MockCloudConnectionManager(
        block=False,
        packages={"package1": {"version": "1.0.0", "supported": True}},
        serverless=False,
    )

    with patch(
        "aikido_zen.helpers.get_machine_ip.get_ip", return_value="192.168.1.1"
    ), patch("platform.python_implementation", return_value="RandomPython"):
        connection_manager_info = get_manager_info(mock_connection_manager)

        assert connection_manager_info["platform"]["name"] == "RandomPython"


def test_get_manager_info_with_different_os():
    mock_connection_manager = MockCloudConnectionManager(
        block=False,
        packages={"package1": {"version": "1.0.0", "supported": True}},
        serverless=False,
    )

    with patch(
        "aikido_zen.helpers.get_machine_ip.get_ip", return_value="192.168.1.1"
    ), patch("platform.python_version", return_value="0.1.2"):
        connection_manager_info = get_manager_info(mock_connection_manager)

        assert connection_manager_info["platform"]["version"] == "0.1.2"


def test_get_manager_info_with_different_os():
    mock_connection_manager = MockCloudConnectionManager(
        block=False,
        packages={"package1": {"version": "1.0.0", "supported": True}},
        serverless=False,
    )

    with patch(
        "aikido_zen.helpers.get_machine_ip.get_ip", return_value="192.168.1.1"
    ), patch("platform.python_version", return_value="0.1.2"), patch(
        "platform.python_implementation", return_value="RandomPython"
    ):
        connection_manager_info = get_manager_info(mock_connection_manager)

        assert connection_manager_info["platform"] == {
            "name": "RandomPython",
            "version": "0.1.2",
        }



import pytest
from unittest.mock import MagicMock
from aikido_zen.background_process.service_config import ServiceConfig
from aikido_zen.ratelimiting.rate_limiter import RateLimiter
from . import should_ratelimit_request


def create_route_metadata(method="POST", route="/login"):
    return {"method": method, "url": route, "route": route}


@pytest.fixture
def user():
    """Fixture to create a mock user."""
    return {"id": "user123"}


def create_connection_manager(endpoints=[], bypassed_ips=[]):
    cm = MagicMock()
    cm.conf = ServiceConfig(
        endpoints=endpoints,
        last_updated_at=1,
        blocked_uids=[],
        bypassed_ips=bypassed_ips,
        received_any_stats=True,
    )
    cm.rate_limiter = RateLimiter(
        max_items=5000, time_to_live_in_ms=120 * 60 * 1000  # 120 minutes
    )
    return cm


def test_rate_limits_by_ip():
    cm = create_connection_manager(
        [
            {
                "method": "POST",
                "route": "/login",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 3,
                    "windowSizeInMS": 1000,
                },
            },
        ]
    )
    route_metadata = create_route_metadata()
    assert should_ratelimit_request(route_metadata, "1.2.3.4", None, cm) == {
        "block": False
    }
    assert should_ratelimit_request(route_metadata, "1.2.3.4", None, cm) == {
        "block": False
    }
    assert should_ratelimit_request(route_metadata, "1.2.3.4", None, cm) == {
        "block": False
    }
    assert should_ratelimit_request(route_metadata, "1.2.3.4", None, cm) == {
        "block": True,
        "trigger": "ip",
    }


def test_rate_limiting_ip_allowed():
    cm = create_connection_manager(
        [
            {
                "method": "POST",
                "route": "/login",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 3,
                    "windowSizeInMS": 1000,
                },
            },
        ],
        ["1.2.3.4"],
    )

    # Test requests from allowed IP
    route_metadata = create_route_metadata()
    assert should_ratelimit_request(route_metadata, "1.2.3.4", None, cm) == {
        "block": False
    }
    assert should_ratelimit_request(route_metadata, "1.2.3.4", None, cm) == {
        "block": False
    }
    assert should_ratelimit_request(route_metadata, "1.2.3.4", None, cm) == {
        "block": False
    }
    assert should_ratelimit_request(route_metadata, "1.2.3.4", None, cm) == {
        "block": False
    }


def test_rate_limiting_by_user(user):
    cm = create_connection_manager(
        [
            {
                "method": "POST",
                "route": "/login",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 3,
                    "windowSizeInMS": 1000,
                },
            },
        ]
    )

    # Test requests from the same user
    route_metadata = create_route_metadata()
    assert should_ratelimit_request(route_metadata, "1.2.3.4", user, cm) == {
        "block": False
    }
    assert should_ratelimit_request(route_metadata, "1.2.3.5", user, cm) == {
        "block": False
    }
    assert should_ratelimit_request(route_metadata, "1.2.3.6", user, cm) == {
        "block": False
    }
    assert should_ratelimit_request(route_metadata, "1.2.3.7", user, cm) == {
        "block": True,
        "trigger": "user",
    }


def test_rate_limiting_with_wildcard():
    cm = create_connection_manager(
        [
            {
                "method": "POST",
                "route": "/api/*",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 3,
                    "windowSizeInMS": 1000,
                },
            },
        ]
    )

    # Test requests to different API endpoints
    assert should_ratelimit_request(
        create_route_metadata(route="/api/login"), "1.2.3.4", None, cm
    ) == {"block": False}
    assert should_ratelimit_request(
        create_route_metadata(route="/api/logout"), "1.2.3.4", None, cm
    ) == {"block": False}
    assert should_ratelimit_request(
        create_route_metadata(route="/api/reset-password"), "1.2.3.4", None, cm
    ) == {"block": False}

    # This request should trigger the rate limit
    assert should_ratelimit_request(
        create_route_metadata(route="/api/login"), "1.2.3.4", None, cm
    ) == {"block": True, "trigger": "ip"}


def test_rate_limiting_with_wildcard2():
    cm = create_connection_manager(
        [
            {
                "method": "*",
                "route": "/api/*",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 3,
                    "windowSizeInMS": 1000,
                },
            },
        ]
    )

    # Test requests to different API endpoints with various methods
    metadata = create_route_metadata(route="/api/login", method="POST")
    assert should_ratelimit_request(metadata, "1.2.3.4", None, cm) == {"block": False}

    metadata = create_route_metadata(route="/api/logout", method="GET")
    assert should_ratelimit_request(metadata, "1.2.3.4", None, cm) == {"block": False}

    metadata = create_route_metadata(route="/api/reset-password", method="PUT")
    assert should_ratelimit_request(metadata, "1.2.3.4", None, cm) == {"block": False}

    # This request should trigger the rate limit
    metadata = create_route_metadata(route="/api/login", method="GET")
    assert should_ratelimit_request(metadata, "1.2.3.4", None, cm) == {
        "block": True,
        "trigger": "ip",
    }


def test_rate_limiting_by_user_with_same_ip():
    cm = create_connection_manager(
        [
            {
                "method": "POST",
                "route": "/login",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 3,
                    "windowSizeInMS": 1000,
                },
            },
        ]
    )

    # First three requests should not be blocked
    metadata = create_route_metadata(route="/login", method="POST")
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }

    # This request should trigger the rate limit
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": True,
        "trigger": "user",
    }


def test_rate_limiting_by_user_with_different_ips():
    cm = create_connection_manager(
        [
            {
                "method": "POST",
                "route": "/login",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 3,
                    "windowSizeInMS": 1000,
                },
            },
        ]
    )

    # First request from first IP
    metadata = create_route_metadata(route="/login", method="POST")
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }

    # First request from second IP
    assert should_ratelimit_request(metadata, "4.3.2.1", {"id": "123"}, cm) == {
        "block": False
    }

    # Second request from first IP
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }

    # This request from second IP should trigger the rate limit
    assert should_ratelimit_request(metadata, "4.3.2.1", {"id": "123"}, cm) == {
        "block": True,
        "trigger": "user",
    }


def test_rate_limiting_same_ip_different_users():
    cm = create_connection_manager(
        [
            {
                "method": "POST",
                "route": "/login",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 3,
                    "windowSizeInMS": 1000,
                },
            },
        ]
    )

    # First request from user 1
    metadata = create_route_metadata(route="/login", method="POST")
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }

    # First request from user 2
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123456"}, cm) == {
        "block": False
    }

    # Second request from user 1
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }

    # Second request from user 2
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123456"}, cm) == {
        "block": False
    }


def test_does_not_ratelimit_bypassed_ip_with_user():
    pass  # Really?


def test_works_with_setuser_after_first_ratelimit():
    pass


import pytest


def test_rate_limiting_bypassed_ip_with_user():
    cm = create_connection_manager(
        [
            {
                "method": "POST",
                "route": "/login",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 3,
                    "windowSizeInMS": 1000,
                },
            },
        ],
        ["1.2.3.4"],
    )

    # All requests from the bypassed IP should not be blocked
    metadata = create_route_metadata(route="/login", method="POST")
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }
    assert should_ratelimit_request(metadata, "1.2.3.4", {"id": "123"}, cm) == {
        "block": False
    }



import pytest
from .check_context_for_path_traversal import check_context_for_path_traversal


# Sample unsafe context for testing
class UnsafeContext:
    cookies = {}
    headers = {}
    remote_address = "1.1.1.1"
    method = "POST"
    url = "http://localhost:8080/"
    query = {}
    body = {"path": "../file"}
    source = "flask"
    route = None


class SafeContext:
    url = "/_next/static/RjAvHy_jB1ciRT_xBrSyI/_ssgManifest.js"
    method = "GET"
    headers = {
        "host": "localhost:3000",
        "connection": "keep-alive",
        "pragma": "no-cache",
        "cache-control": "no-cache",
        "sec-ch-ua": '"Chromium";v="124", "Google Chrome";v="124", "Not-A.Brand";v="99"',
        "sec-ch-ua-mobile": "?0",
        "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
        "sec-ch-ua-platform": '"macOS"',
        "accept": "*/*",
        "sec-fetch-site": "same-origin",
        "sec-fetch-mode": "no-cors",
        "sec-fetch-dest": "script",
        "referer": "http://localhost:3000/",
        "accept-encoding": "gzip, deflate, br, zstd",
        "accept-language": "nl,en;q=0.9,en-US;q=0.8",
        "cookie": "Phpstorm-8262f4a6=6a1925f9-2f0e-45ea-8336-a6988d56b1aa",
        "x-forwarded-host": "localhost:3000",
        "x-forwarded-port": "3000",
        "x-forwarded-proto": "http",
        "x-forwarded-for": "127.0.0.1",
    }
    route = None
    query = {}
    source = "flask"
    cookies = {
        "Phpstorm-8262f4a6": "6a1925f9-2f0e-45ea-8336-a6988d56b1aa",
    }
    body = None
    remote_address = "127.0.0.1"


def test_detects_path_traversal_from_route_parameter(monkeypatch):
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)
    result = check_context_for_path_traversal("../file/test.txt", "op", UnsafeContext())
    assert result == {
        "operation": "op",
        "kind": "path_traversal",
        "source": "body",
        "pathToPayload": ".path",
        "metadata": {
            "filename": "../file/test.txt",
        },
        "payload": "../file",
    }


def test_does_not_flag_safe_operation(monkeypatch):
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    filename = "../../web/spec-extension/cookies"
    operation = "path.normalize"
    result = check_context_for_path_traversal(filename, operation, SafeContext())
    assert len(result) == 0


def test_detects_path_traversal_with_buffer(monkeypatch):
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    filename = b"../file/test.txt"
    result = check_context_for_path_traversal(filename, "op", UnsafeContext())
    assert result == {
        "operation": "op",
        "kind": "path_traversal",
        "source": "body",
        "pathToPayload": ".path",
        "metadata": {
            "filename": "../file/test.txt",
        },
        "payload": "../file",
    }


def test_detects_path_traversal_with_url(monkeypatch):
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    filename = "file:///../file/test.txt"
    result = check_context_for_path_traversal(filename, "op", UnsafeContext())
    assert result == {
        "operation": "op",
        "kind": "path_traversal",
        "source": "body",
        "pathToPayload": ".path",
        "metadata": {
            "filename": "/../file/test.txt",
        },
        "payload": "../file",
    }


def test_ignores_non_utf8_buffer(monkeypatch):
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    filename = bytes([0x80, 0x81, 0x82, 0x83])
    result = check_context_for_path_traversal(filename, "op", UnsafeContext())
    assert len(result) == 0


def test_ignores_invalid_filename_type(monkeypatch):
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    filename = object()
    result = check_context_for_path_traversal(filename, "op", UnsafeContext())
    assert len(result) == 0
